# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 21:55
# @File    : list.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function



class ListMixin:
    """
    Mixin to make dc an extension of python list.
    """
    def append(self, *args) -> 'DataCollection':
        """
        Append item to data collection.

        Args:
            item (Any): the item to append

        Returns:
            DataCollection: self

        Examples:

        >>> dc = DataCollection([0, 1, 2])
        >>> dc.append(3).append(4)
        [0, 1, 2, 3, 4]
        """
        if hasattr(self._iterable, 'append'):
            self._iterable.append(*args)
            return self
        raise TypeError('append() is only supported for data collection created from list.')

    def clear(self, *args) -> 'DataCollection':
        """
        Clear a DataCollection.

        Examples:

        >>> dc = DataCollection([1, 2, 3])
        >>> dc.clear()
        []
        """
        if hasattr(self._iterable, 'clear'):
            self._iterable.clear(*args)
            return self
        raise TypeError('clear() is only supported for data collection created from list.')

    def copy(self, *args) -> 'DataCollection':
        """
        Copy a DataCollection.

        Examples:

        >>> dc = DataCollection([1, 2, 3])
        >>> dc_1 = dc.copy()
        >>> dc_1._iterable.append(4)
        >>> dc, dc_1
        ([1, 2, 3], [1, 2, 3, 4])
        """
        if hasattr(self._iterable, 'copy'):
            return self._factory(self._iterable.copy(*args))
        raise TypeError('copy() is only supported for data collection created from list.')

    def count(self, *args) -> int:
        """
        Count an element in  DataCollection.

        Examples:

        >>> dc = DataCollection([1, 2, 3])
        >>> dc.count(1)
        1
        """
        if hasattr(self._iterable, 'count'):
            return self._iterable.count(*args)
        raise TypeError('count() is only supported for data collection created from list.')

    def extend(self, *args) -> 'DataCollection':
        """
        Extend a DataCollection.

        Examples:

        >>> dc = DataCollection([1, 2, 3])
        >>> dc.extend([4, 5])
        [1, 2, 3, 4, 5]
        """
        if hasattr(self._iterable, 'extend'):
            self._iterable.extend(*args)
            return self
        raise TypeError('extend() is only supported for data collection created from list.')

    def insert(self, *args) -> 'DataCollection':
        """
        Insert data into a DataCollection.

        Examples:

        >>> dc = DataCollection([1, 2, 3])
        >>> dc.insert(0, 0)
        [0, 1, 2, 3]
        """
        if hasattr(self._iterable, 'insert'):
            self._iterable.insert(*args)
            return self
        raise TypeError('insert() is only supported for data collection created from list.')

    def pop(self, *args) -> 'DataCollection':
        """
        Extend a DataCollection.

        Examples:

        >>> dc = DataCollection([1, 2, 3])
        >>> dc.pop()
        [1, 2]
        """
        if hasattr(self._iterable, 'pop'):
            self._iterable.pop(*args)
            return self
        raise TypeError('pop() is only supported for data collection created from list.')

    def remove(self, *args) -> 'DataCollection':
        """
        Remove element from DataCollection.

        Examples:

        >>> dc = DataCollection([1, 2, 3])
        >>> dc.remove(1)
        [2, 3]
        """
        if hasattr(self._iterable, 'remove'):
            self._iterable.remove(*args)
            return self
        raise TypeError('remove() is only supported for data collection created from list.')

    def reverse(self, *args) -> 'DataCollection':
        """
        Reverse a DataCollection.

        Examples:

        >>> dc = DataCollection([1, 2, 3])
        >>> dc.reverse()
        [3, 2, 1]
        """
        if hasattr(self._iterable, 'reverse'):
            self._iterable.reverse(*args)
            return self
        raise TypeError('reverse() is only supported for data collection created from list.')

    def sort(self, *args) -> 'DataCollection':
        """
        Sort a DataCollection.

        Examples:

        >>> dc = DataCollection([1, 4, 3])
        >>> dc.sort()
        [1, 3, 4]
        """
        if hasattr(self._iterable, 'sort'):
            self._iterable.sort(*args)
            return self
        raise TypeError('sort() is only supported for data collection created from list.')

    # def __len__(self):
    #     """
    #     Return the length of iterable.
    #     """
    #     if not self.is_stream():
    #         return self._iterable.__len__()
    #     else:
    #         raise TypeError('Streamed data collection does not support len.')



from typing import List, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from antgo.framework.helper.models.builder import HEADS, MODELS
from antgo.framework.helper.runner import BaseModule


@MODELS.register_module()
class EncoderDecoder(BaseModule):
    def __init__(self,
                 backbone,
                 decode_head,
                 neck = None,
                 train_cfg = None,
                 test_cfg = None,
                 init_cfg = None):
        super().__init__(init_cfg=init_cfg)
        # if pretrained is not None:
        #     assert backbone.get('pretrained') is None, \
        #         'both backbone and segmentor set pretrained weight'
        #     backbone.pretrained = pretrained
        self.backbone = MODELS.build(backbone)
        self.with_neck = False
        if neck is not None:
            self.neck = MODELS.build(neck)
            self.with_neck = True

        self.decode_head = HEADS.build(decode_head)
        self.train_cfg = train_cfg
        self.test_cfg = test_cfg

    def extract_feat(self, image):
        """Extract features from images."""
        x = self.backbone(image)
        if self.with_neck:
            x = self.neck(x)
        return x

    def forward_train(self, image, segments, **kwargs):
        x = self.extract_feat(image)
        x = self.decode_head(x)

        losses = dict()
        loss_decode = \
            self.decode_head.loss(x, segments.long())
        losses.update(loss_decode)
        return losses

    def forward_test(self, image, **kwargs):
        x = self.extract_feat(image)
        seg_predict = self.decode_head.simple_test(x)
        seg_predict = F.interpolate(seg_predict, size=(image.shape[2:]), mode='bilinear')
        seg_predict = torch.argmax(seg_predict, 1)
        return {
            'pred_segments': seg_predict
        }

    def onnx_export(self, image):
        feat = self.extract_feat(image)
        seg_predict = self.decode_head.simple_test(feat)
        return seg_predict



# Copyright (c) OpenMMLab. All rights reserved.
from ...parallel import is_module_wrapper
from ..hooks.hook import HOOKS, Hook


@HOOKS.register_module()
class EMAHook(Hook):
    r"""Exponential Moving Average Hook.

    Use Exponential Moving Average on all parameters of model in training
    process. All parameters have a ema backup, which update by the formula
    as below. EMAHook takes priority over EvalHook and CheckpointSaverHook.

        .. math::

            Xema\_{t+1} = (1 - \text{momentum}) \times
            Xema\_{t} +  \text{momentum} \times X_t

    Args:
        momentum (float): The momentum used for updating ema parameter.
            Defaults to 0.0002.
        interval (int): Update ema parameter every interval iteration.
            Defaults to 1.
        warm_up (int): During first warm_up steps, we may use smaller momentum
            to update ema parameters more slowly. Defaults to 100.
        resume_from (str): The checkpoint path. Defaults to None.
    """

    def __init__(self,
                 momentum=0.0002,
                 interval=1,
                 warm_up=100,
                 resume_from=None):
        assert isinstance(interval, int) and interval > 0
        self.warm_up = warm_up
        self.interval = interval
        assert momentum > 0 and momentum < 1
        self.momentum = momentum**interval
        self.checkpoint = resume_from

    def before_run(self, runner):
        """To resume model with it's ema parameters more friendly.

        Register ema parameter as ``named_buffer`` to model
        """
        model = runner.model
        if is_module_wrapper(model):
            model = model.module
        self.param_ema_buffer = {}
        self.model_parameters = dict(model.named_parameters(recurse=True))
        for name, value in self.model_parameters.items():
            # "." is not allowed in module's buffer name
            buffer_name = f"ema_{name.replace('.', '_')}"
            self.param_ema_buffer[name] = buffer_name
            model.register_buffer(buffer_name, value.data.clone())
        self.model_buffers = dict(model.named_buffers(recurse=True))
        if self.checkpoint is not None:
            runner.resume(self.checkpoint)

    def after_train_iter(self, runner):
        """Update ema parameter every self.interval iterations."""
        curr_step = runner.iter
        # We warm up the momentum considering the instability at beginning
        momentum = min(self.momentum,
                       (1 + curr_step) / (self.warm_up + curr_step))
        if curr_step % self.interval != 0:
            return
        for name, parameter in self.model_parameters.items():
            buffer_name = self.param_ema_buffer[name]
            buffer_parameter = self.model_buffers[buffer_name]
            buffer_parameter.mul_(1 - momentum).add_(momentum, parameter.data)

    def after_train_epoch(self, runner):
        """We load parameter values from ema backup to model before the
        EvalHook."""
        self._swap_ema_parameters()

    def before_train_epoch(self, runner):
        """We recover model's parameter from ema backup after last epoch's
        EvalHook."""
        self._swap_ema_parameters()

    def _swap_ema_parameters(self):
        """Swap the parameter of model with parameter in ema_buffer."""
        for name, value in self.model_parameters.items():
            temp = value.data.clone()
            ema_buffer = self.model_buffers[self.param_ema_buffer[name]]
            value.data.copy_(ema_buffer.data)
            ema_buffer.data.copy_(temp)



# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
import copy
from functools import reduce
from functools import partial


def add_tensor_function(func):
    setattr(paddle.Tensor, func.__name__, func)


@property
def data(self):
    return self


setattr(paddle.Tensor, "data", data)


@property
def requires_grad(self):
    return not self.stop_gradient


setattr(paddle.Tensor, "requires_grad", requires_grad)


@add_tensor_function
def requires_grad_(self, requires_grad=True):
    self.stop_gradient = not requires_grad
    return self


@add_tensor_function
def item(self):
    return self.numpy()[0]


@add_tensor_function
def permute(self, *dims):
    return self.transpose(dims)


@add_tensor_function
def clamp_(self, min, max):
    return self.clip(min, max)


@add_tensor_function
def contiguous(self):
    return self


@add_tensor_function
def view(self, *shape):
    return self.reshape(*shape)


@add_tensor_function
def repeat(self, *sizes):
    return self.tile(sizes)


@add_tensor_function
def dim(self):
    return self.ndim


@add_tensor_function
def long(self, memory_format=None):
    return paddle.cast(self, dtype="int64")


@add_tensor_function
def float(self, memory_format=None):
    return paddle.cast(self, dtype="float32")


@add_tensor_function
def size(self, dim=None):
    if dim is not None:
        return self.shape[dim]
    else:
        return self.shape


@add_tensor_function
def to(self, *args, **kwargs):
    if len(args) == 1 and "dtype" not in kwargs:
        try:
            return paddle.cast(self, dtype=args[0])
        except Exception:
            return self
    else:
        if len(kwargs) > 0:
            if "dtype" in kwargs:
                return paddle.cast(self, dtype=kwargs["dtype"])
            else:
                return self
        else:
            return self


@add_tensor_function
def index_fill_(self, dim, index, val):
    x_shape = self.shape
    index_shape = index.shape
    if dim != 0:
        perm_list = list(range(len(x_shape)))
        while dim < 0:
            dim += len(x_shape)
        perm_list.pop(dim)
        perm_list = [dim] + perm_list
        self = paddle.transpose(self, perm=perm_list)
        s = x_shape.pop(dim)
        x_shape = [s] + x_shape
    updates_shape = index_shape + x_shape[1:]
    updates = paddle.full(updates_shape, fill_value=val, dtype=self.dtype)
    out = paddle.scatter(self, index, updates)
    if dim != 0:
        perm_list = list(range(len(x_shape)))
        perm_list.pop(0)
        perm_list.insert(dim, 0)
        out = paddle.transpose(out, perm=perm_list)
    paddle.assign(out, output=self)


@add_tensor_function
def fill_(self, value):
    paddle.assign(
        paddle.full_like(
            self, value, dtype="float32").cast(self.dtype),
        output=self)


pd_sum = partial(paddle.Tensor.sum)


@add_tensor_function
def sum(self, dim, keepdim=False, dtype=None):
    return pd_sum(self, axis=dim, dtype=dtype, keepdim=keepdim)


pd_sort = partial(paddle.Tensor.sort)


@add_tensor_function
def sort(self, dim=-1, descending=False, out=None):
    return pd_sort(
        self, axis=dim, descending=descending), paddle.argsort(
            self, axis=dim, descending=descending)


pd_reshape = partial(paddle.Tensor.reshape)


@add_tensor_function
def reshape(self, *shape):
    return pd_reshape(self, shape)


pd_transpose = partial(paddle.Tensor.transpose)


@add_tensor_function
def transpose(self, dim0, dim1=None):
    if dim1 is None:
        return pd_transpose(self, dim0)
    else:
        shape = self.shape
        perm = list(range(len(shape)))
        dim0 = (dim0 + len(shape)) if dim0 < 0 else dim0
        dim1 = (dim1 + len(shape)) if dim1 < 0 else dim1
        perm[dim0] = dim1
        perm[dim1] = dim0
        return pd_transpose(self, perm)


pd_max = partial(paddle.Tensor.max)


@add_tensor_function
def max(self, dim, keepdim=None):
    return pd_max(self, dim, keepdim), paddle.argmax(self, dim, keepdim)


pd_min = partial(paddle.Tensor.min)


@add_tensor_function
def min(self, dim, keepdim=None):
    return pd_min(self, dim, keepdim), paddle.argmin(self, dim, keepdim)


pd_expand = partial(paddle.Tensor.expand)


@add_tensor_function
def expand(self, *sizes):
    return pd_expand(self, sizes)


@add_tensor_function
def div(self, value):
    return self / value


@add_tensor_function
def eq(self, other):
    return self.equal(other)


@add_tensor_function
def eq_(self, other):
    return self.equal(other)


@add_tensor_function
def mul(self, value):
    return self * value


@add_tensor_function
def mul_(self, value):
    return self * value


pd_cuda = partial(paddle.Tensor.cuda)


@add_tensor_function
def cuda(self, device=None, non_blocking=False, memory_format=None):
    return self


@add_tensor_function
def copy_(self, src, non_blocking=False):
    src = paddle.expand(src, self.shape)
    paddle.assign(src, self)



import numpy as np
import torch
from antgo.framework.helper.dataset.builder import DATASETS
from antgo.framework.helper.reader import *
from antgo.ant import environment
import multiprocessing as mp
import pickle


def get_keys(args):
    return environment.KVReader(*args).list_keys()


def worker_init_fn(path, dataset, _):
    dataset.reader = environment.KVReader(path, 2)


@DATASETS.register_module()
class KVDatasetReader(KVReaderBase):
    def __init__(self, data_path_list, pipeline=None, weak_pipeline=None, strong_pipeline=None) -> None:
        super().__init__(pipeline, weak_pipeline, strong_pipeline)
        self.data_path_list = data_path_list

        # TODO, 支持多数据组合

        # 读取所有可以
        with mp.Pool(1) as p:
            self.keys = p.map(get_keys, [(self.data_path_list[0], 1)])[0]

        self.worker_init_fn = lambda worker_id: worker_init_fn(self.data_path_list[0], self, worker_id)

    def reads(self, index):
        if not getattr(self, 'reader', None):
            self.worker_init_fn(0)
            
        # index 是一个list
        index = [self.keys[i] for i in index]
        data_bytes_list = self.reader.read_many(index)

        samples = []
        for data_bytes in data_bytes_list:
            data = pickle.loads(data_bytes)
            samples.append(data)
        
        return samples



# Once for All: Train One Network and Specialize it for Efficient Deployment
# Han Cai, Chuang Gan, Tianzhe Wang, Zhekai Zhang, Song Han
# International Conference on Learning Representations (ICLR), 2020.

import json
import torch

from antgo.framework.helper.models.nas.utils import download_url
# from ofa.imagenet_classification.networks import get_net_by_name, proxyless_base
# from ofa.imagenet_classification.elastic_nn.networks import (
#     OFAMobileNetV3,
#     OFAProxylessNASNets,
#     OFAResNets,
# )
from antgo.framework.helper.models.nas.ofa.networks.ofa_mbv3 import OFAMobileNetV3

__all__ = [
    "ofa_specialized",
    "ofa_net",
    "proxylessnas_net",
    "proxylessnas_mobile",
    "proxylessnas_cpu",
    "proxylessnas_gpu",
]


def ofa_specialized(net_id, pretrained=True):
    url_base = "https://hanlab.mit.edu/files/OnceForAll/ofa_specialized/"
    net_config = json.load(
        open(
            download_url(
                url_base + net_id + "/net.config",
                model_dir=".torch/ofa_specialized/%s/" % net_id,
            )
        )
    )
    net = get_net_by_name(net_config["name"]).build_from_config(net_config)

    image_size = json.load(
        open(
            download_url(
                url_base + net_id + "/run.config",
                model_dir=".torch/ofa_specialized/%s/" % net_id,
            )
        )
    )["image_size"]

    if pretrained:
        init = torch.load(
            download_url(
                url_base + net_id + "/init",
                model_dir=".torch/ofa_specialized/%s/" % net_id,
            ),
            map_location="cpu",
        )["state_dict"]
        net.load_state_dict(init)
    return net, image_size


def ofa_net(net_id, pretrained=True):
    if net_id == "ofa_proxyless_d234_e346_k357_w1.3":
        net = OFAProxylessNASNets(
            dropout_rate=0,
            width_mult=1.3,
            ks_list=[3, 5, 7],
            expand_ratio_list=[3, 4, 6],
            depth_list=[2, 3, 4],
        )
    elif net_id == "ofa_mbv3_d234_e346_k357_w1.0":
        net = OFAMobileNetV3(
            dropout_rate=0,
            width_mult=1.0,
            ks_list=[3, 5, 7],
            expand_ratio_list=[3, 4, 6],
            depth_list=[2, 3, 4],
        )
    elif net_id == "ofa_mbv3_d234_e346_k357_w1.2":
        net = OFAMobileNetV3(
            train_cfg=dict(
                dropout_rate=0,
                width_mult=1.2,
                ks_list=[3, 5, 7],
                expand_ratio_list=[3, 4, 6],
                depth_list=[2, 3, 4],
            )
        )
    elif net_id == "ofa_resnet50":
        net = OFAResNets(
            dropout_rate=0,
            depth_list=[0, 1, 2],
            expand_ratio_list=[0.2, 0.25, 0.35],
            width_mult_list=[0.65, 0.8, 1.0],
        )
        net_id = "ofa_resnet50_d=0+1+2_e=0.2+0.25+0.35_w=0.65+0.8+1.0"
    else:
        raise ValueError("Not supported: %s" % net_id)

    if pretrained:
        url_base = "https://hanlab.mit.edu/files/OnceForAll/ofa_nets/"
        init = torch.load(
            download_url(url_base + net_id, model_dir=".torch/ofa_nets"),
            map_location="cpu",
        )["state_dict"]
        net.load_state_dict(init)
    return net


def proxylessnas_net(net_id, pretrained=True):
    net = proxyless_base(
        net_config="https://hanlab.mit.edu/files/proxylessNAS/%s.config" % net_id,
    )
    if pretrained:
        net.load_state_dict(
            torch.load(
                download_url(
                    "https://hanlab.mit.edu/files/proxylessNAS/%s.pth" % net_id
                ),
                map_location="cpu",
            )["state_dict"]
        )
    return net


def proxylessnas_mobile(pretrained=True):
    return proxylessnas_net("proxyless_mobile", pretrained)


def proxylessnas_cpu(pretrained=True):
    return proxylessnas_net("proxyless_cpu", pretrained)


def proxylessnas_gpu(pretrained=True):
    return proxylessnas_net("proxyless_gpu", pretrained)



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/11 23:02
# @File    : __init__.py.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from .dag import DagMixin
from .config import ConfigMixin
from .column import ColumnMixin
from .dispatcher import DispatcherMixin
from .stream import StreamMixin
from .data_processing import DataProcessingMixin
from .list import ListMixin
from .dataset import DatasetMixin
from .computer_vision import ComputerVisionMixin
from .dataframe import DataFrameMixin
from .show import ShowMixin
from .serve import ServeMixin
from .demo import DemoMixin
from .deploy import DeployMixin

class DCMixins(
  DagMixin,
  ConfigMixin,
  ColumnMixin,
  DispatcherMixin,
  StreamMixin,
  DataProcessingMixin,
  ListMixin,
  DatasetMixin,
  ComputerVisionMixin,
  DataFrameMixin,
  ShowMixin,
  ServeMixin,
  DemoMixin,
  DeployMixin):

  def __init__(self) -> None:  # pylint: disable=useless-super-delegation
    super().__init__()



import warnings

from antgo.framework.helper.cnn import MODELS as CV_MODELS
from antgo.framework.helper.utils import Registry

MODELS = Registry('models', parent=CV_MODELS)

CLASSIFIERS = MODELS
BACKBONES = MODELS
NECKS = MODELS
ROI_EXTRACTORS = MODELS
SHARED_HEADS = MODELS
HEADS = MODELS
LOSSES = MODELS
DETECTORS = MODELS
DISTILLER = MODELS
DISTILL_LOSSES = MODELS


def build_backbone(cfg):
    """Build backbone."""
    return BACKBONES.build(cfg)


def build_neck(cfg):
    """Build neck."""
    return NECKS.build(cfg)


def build_roi_extractor(cfg):
    """Build roi extractor."""
    return ROI_EXTRACTORS.build(cfg)


def build_shared_head(cfg):
    """Build shared head."""
    return SHARED_HEADS.build(cfg)


def build_head(cfg):
    """Build head."""
    return HEADS.build(cfg)


def build_loss(cfg):
    """Build loss."""
    return LOSSES.build(cfg)


def build_model(cfg):
    """Build detector."""
    return MODELS.build(
        cfg, 
        default_args=dict(
            train_cfg=None,
            test_cfg=None
        )
    )


def build_distiller(cfg, teacher_cfg, student_cfg):
    """Build distiller."""
    # if train_cfg is not None or test_cfg is not None:
    #     warnings.warn(
    #         'train_cfg and test_cfg is deprecated, '
    #         'please specify them in model', UserWarning)
    # assert cfg.get('train_cfg') is None or train_cfg is None, \
    #     'train_cfg specified in both outer field and model field '
    # assert cfg.get('test_cfg') is None or test_cfg is None, \
    #     'test_cfg specified in both outer field and model field '
    """Build detector."""
    return MODELS.build(
        cfg, 
        default_args=dict(
            teacher_cfg=teacher_cfg,
            student_cfg=student_cfg
        )
    )    

def build_distill_loss(cfg):
    """Build distill loss."""
    return DISTILL_LOSSES.build(cfg)


# -*- coding: UTF-8 -*-
# @Time    : 2020-06-25 23:30
# @File    : browser_server.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import copy
import numpy as np
import tornado.httpserver
import tornado.ioloop
import tornado.options
from tornado.options import define, options
from tornado import web, gen
from tornado import httpclient
import tornado.web
import os
import shutil
import signal
from antgo.crowdsource.base_server import *
from antgo.utils import logger
from antgo.ant import environment
import traceback
import sys
import uuid
import json
import time
import base64
import requests
import cv2
import torch
mano_layer_map = None


# # 新用户，从队列中获取新数据并加入用户队列中
# self.db['data'].append({
#   'value': self.response_queue.get(),
#   'status': False,
#   'time': time.time()
# })

class FreshApiHandler(BaseHandler):
  @gen.coroutine
  def post(self):
    # 添加样本记录
    samples = self.get_argument('samples', None)
    if samples is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    # 解析json
    samples = json.loads(samples)

    is_replace = self.get_argument('is_replace', None)
    if is_replace:
      self.db['data'] = []

    for sample_i, sample in enumerate(samples):
      for data in sample:
        if data['type'] == 'IMAGE':
          large_size = max(data['width'], data['height'])
          scale = large_size / 400
          if scale > 1.0:
            data['width'] = (int)(data['width'] / scale)
            data['height'] = (int)(data['height'] / scale)

      self.db['data'].append({
        'value': sample,
        'status': False,
        'time': time.time()
      })

    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class EntryApiHandler(BaseHandler):
  @gen.coroutine
  def get(self, *args, **kwargs):
    user = self.get_current_user()
    if user is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    # 检查是否有输入，如果有先下载，在更新数据库
    input_info = self.get_argument('input', '')
    if input_info != '':
      # step 1: 下载 
      if input_info.startswith('http') and input_info.endswith('json'):
        # step 1.1: 下载 (from url)
        try:
          static_path = self.settings.get('static_path')
          if not input_info.endswith('.json'):
            self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
            return
          input_ext_name = input_info.split('.')[-1]          
          local_path = os.path.join(static_path, f'download.{input_ext_name}')
          
          data = requests.get(input_info, timeout=7)
          with open(local_path, 'wb') as fp:
            fp.write(data.content)     
          
          input_info = local_path
        except:
          logger.error('Fail to download from html.')
          self.response(RESPONSE_STATUS_CODE.EXECUTE_FORBIDDEN)
          return
      elif input_info.lower().startswith("hdfs") and input_info.endswith('json'):
        # step 1.2: 下载 (from htfs, jsonfile)
        try:
          static_path = self.settings.get('static_path')
          environment.hdfs_client.get(input_info, static_path)
          file_name = input_info.split('/')[-1]
          local_path = os.path.join(static_path, file_name)
          if not os.path.exists(local_path):
            logger.error('Fail to download from hdfs.')
            self.response(RESPONSE_STATUS_CODE.EXECUTE_FORBIDDEN)
            return
          
          input_info = local_path
        except:
          logger.error('Fail to download from hdfs.')
          self.response(RESPONSE_STATUS_CODE.EXECUTE_FORBIDDEN)
          return
      elif input_info.lower().startswith("hdfs"):
        # step 1.3: 下载（from htfs, folder)
        try:
          static_path = self.settings.get('static_path')          
          sub_folder = 'json-folder'
          if os.path.exists(os.path.join(static_path, sub_folder)):
            shutil.rmtree(os.path.join(static_path, sub_folder))
          os.makedirs(os.path.join(static_path, sub_folder))
          
          temp_path = os.path.join(static_path, sub_folder)
          if input_info.endswith('/'):
            input_info = input_info[:-1]            
          environment.hdfs_client.get(f'{input_info}/*', temp_path)

          json_file_path = ''
          if os.path.isdir(temp_path):
            for file_name in os.listdir(temp_path):
              if file_name.endswith('json'):
                json_file_path = os.path.join(temp_path, file_name)
                shutil.copyfile(json_file_path, os.path.join(static_path, file_name))
                json_file_path = os.path.join(static_path, file_name)
                break
          
          assert(json_file_path != '')
          input_info = json_file_path
        except:
          logger.error('Fail to download folder from hdfs.')
          self.response(RESPONSE_STATUS_CODE.EXECUTE_FORBIDDEN)
          return
      else:
        # step 1.4: 本地路径
        if not os.path.exists(input_info):
          self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
          return
        
      # step 2: 更新数据库
      self.update_db(self.settings.get('static_path'), input_info)
    
    if len(self.db['data']) == 0:
      # 当前无数据，直接返回
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
      return

    user_name = user['full_name']
    if user_name not in self.db['user_record']:
      self.db['user_record'][user_name] = []
    if len(self.db['user_record'][user_name]) == 0:
      self.db['user_record'][user_name].append(0)

    # 获得当前用户浏览位置
    entry_id = self.db['user_record'][user_name][-1]

    # 构建返回数据
    try:
      response_content = {
        'value': update_vis_elem(self.db['data'][entry_id]['value']),
        'step': len(self.db['user_record'][user_name]) - 1,
        'tags': self.settings.get('tags', []),
        'operators': [],
        'dataset_flag': self.db['state'],
        'samples_num': self.db['dataset'][self.db['state']]['samples_num'],
        'samples_num_checked': self.db['dataset'][self.db['state']]['samples_num_checked'],
      }
    except Exception:
      print(traceback.format_exc())
      self.response(RESPONSE_STATUS_CODE.INTERNAL_SERVER_ERROR)
      return

    self.response(RESPONSE_STATUS_CODE.SUCCESS, content=response_content)
    return

  def update_db(self, dump_dir, json_file_path):
    # 1.step 清空数据
    self.db['data'] = []
    
    # 2.step 解析下载数据（json, or tar）
    if not os.path.exists(json_file_path):
      return
    if os.path.isdir(json_file_path):
      for file_name in os.listdir(json_file_path):
        if file_name.endswith('json'):
          json_file_path = os.path.join(json_file_path, file_name)
          break
              
    sample_anno_json_file = json_file_path
    sample_list = []
    # step2.1: 加载样本信息
    # 兼容两种常用的存储样本信息方式（1.纯json格式，所有样本以list形式存储；2.样本按行存储，每个样本的信息是json格式）
    try:
      # try 1 (纯json格式，所有样本以list形式存储)
      with open(sample_anno_json_file, 'r', encoding="utf-8") as fp:
        sample_list = json.load(fp)
    except:
      # try 2 (样本信息按行存储，每个样本信息是json格式)
      with open(sample_anno_json_file, 'r', encoding="utf-8") as fp:
        sample_info_str = fp.readline()
        sample_info_str = sample_info_str.strip()
        
        while sample_info_str:
          sample_info = json.loads(sample_info_str)
          sample_list.append(sample_info)
          sample_info_str = fp.readline()
          sample_info_str = sample_info_str.strip()
          if sample_info_str == '':
            break        
          
    # step2.2: 尝试加载样本集meta信息
    sample_meta = {'meta': {}}
    sample_folder = os.path.dirname(sample_anno_json_file)
    data_meta_file = os.path.join(sample_folder, 'meta.json')
    if os.path.exists(data_meta_file):
      with open(data_meta_file, 'r', encoding="utf-8") as fp:
        sample_meta = json.load(fp)
            
    # 3.step 加载数据
    sample_folder = os.path.abspath(sample_folder)  # 将传入的路径转换为绝对路径
    static_dir = self.settings.get('static_path')
    if sample_list is not None and sample_folder is not None:
      # 为样本所在目录建立软连接到static下面
      if not sample_folder.startswith(static_dir):
        # 建立软连接
        os.system(f'cd {static_dir}; ln -s {sample_folder} dataset;')
      
      # 将数据信息写如本地数据库
      # 1.step 尝试加载meta信息
      meta_info = sample_meta
      # 2.step 样本信息写入数据库
      for sample_id, sample in enumerate(sample_list):
        # 仅考虑image_url
        data_source = sample['image_url']
        file_name = data_source
        sample_label = ''
        if 'image_label_name' in sample:
          sample_label = sample['image_label_name']
        elif 'image_label' in sample:
          sample_label = sample['image_label']
 
        convert_sample = [{
          'type': 'IMAGE',
          'data': data_source,
          'tag': [sample_label],
          'title': sample['image_url'],
          'id': sample_id
        }]
        
        #############################  标准信息  ###############################
        # 添加框元素
        if 'bboxes' in sample:
          convert_sample[0].update({
            'bboxes': sample['bboxes']
          })
        # 添加多边形元素
        if 'segments' in sample:
            convert_sample[0].update({
            'segments': sample['segments']
          })          
        # 添加框标签信息
        if 'labels' in sample:
          convert_sample[0].update({
            'labels': sample['labels']
          })
        # 添加样本标签
        if 'image_label' in sample:
            convert_sample[0].update({
            'image_label': sample['image_label']
          })
            
        #############################  扩展信息 1  #############################
        # 添加2d关键点元素
        if 'joints2d' in sample:
          convert_sample[0].update({
            'joints2d': sample['joints2d'],
            'skeleton': meta_info['meta']['skeleton'] if 'skeleton' in meta_info['meta'] else []
          })
        
        #############################  扩展信息 2  #############################
        # 添加3d关键点元素（需要相机参数）
        if 'joints3d' in sample and \
          'cam_param' in sample and len(sample['cam_param']) > 0:
          convert_sample.append({
            'type': 'IMAGE',
            'data': data_source,
            'tag': [sample_label],
            'title': f'3D-POINTS-({file_name})',
            'id': sample_id
          })
          convert_sample[-1].update({
            'joints3d': sample['joints3d'],
            'skeleton': meta_info['meta']['skeleton'] if 'skeleton' in meta_info['meta'] else [],
            'cam_param': sample['cam_param']
          })
        
        #############################  扩展信息 3  #############################    
        # 添加3d关键点元素（mano）(需要相机参数)
        if 'pose' in sample and 'trans' in sample and 'shape' in sample and \
          'cam_param' in sample and len(sample['cam_param']) > 0 and \
          'model' in meta_info['meta']:
          
          pose_shape_model = meta_info['meta']['model']
          convert_sample.append({
            'type': 'IMAGE',
            'data': data_source,
            'tag': [sample_label],
            'title': f'{pose_shape_model}-({file_name})',
            'id': sample_id
          })
          convert_sample[-1].update({
            'skeleton': meta_info['meta']['skeleton'] if 'skeleton' in meta_info['meta'] else [],
            'labels': sample['labels'] if 'labels' in sample else [],
            'category': meta_info['meta']['category'] if 'category' in meta_info['meta'] else {},            
            'pose': sample['pose'],
            'shape': sample['shape'],
            'trans': sample['trans'],
            'cam_param': sample['cam_param'],
            'model': pose_shape_model
          })          
                
        self.db['data'].append({
          'value': convert_sample,
          'status': False,
          'time': time.time()
        })


def update_vis_elem(data_group):
  data_group = copy.deepcopy(data_group)
  for data in data_group:
    if 'joints3d' in data and 'cam_param' in data:
      # 基于相机模型，进行投影
      if 'D' in data['cam_param'] and 'Xi' in data['cam_param'] and 'K' in data['cam_param']:
        # 鱼眼模型(omni)
        # points_3d: Nx21x3
        K = np.array(data['cam_param']['K']).astype(np.float32)
        Xi = float(data['cam_param']['Xi'])
        D = np.array(data['cam_param']['D']).astype(np.float32)
        points_2d_list = []
        for i in range(len(data['joints3d'])):
          points_3d = np.array(data['joints3d'][i]).astype(np.float32)
          points_3d = points_3d.reshape(points_3d.shape[0],1,3).astype(np.float32)
          uv,_ = cv2.omnidir.projectPoints(points_3d, np.zeros(3), np.zeros(3), K, Xi, D)
          points_2d = uv[:,0,:].tolist()
          points_2d_list.append(points_2d)
        
        # update elem
        data.update({
          'joints2d': points_2d_list
        })
      elif 'K' in data['cam_param']:
        # 小孔成像模型
        # points_3d: Nx21x3
        K = np.array(data['cam_param']['K']).astype(np.float32)
        points_2d_list = []
        for i in range(len(data['joints3d'])):
          points_3d = np.array(data['joints3d'][i])
          points_2d = np.matmul(K, np.transpose(points_3d))
          points_2d = np.transpose(points_2d).tolist()
          points_2d_list.append(points_2d)
        
        # update elem
        data.update({
          'joints2d': points_2d_list
        })
    elif 'model' in data and 'cam_param' in data:
      # 基于物体模型（mano,smpl）->绘制3D点->绘制2D点
      global mano_layer_map
      if data['model'] == 'mano' and mano_layer_map is not None:
        try:
          category_names = []
          if len(data['labels']) > 0 and len(data['category']) > 0:
            # left or right
            category_names = [data['category'][str(label)] for label in data['labels']]
                    
          points_2d_list = []
          for i in range(len(data['pose'])):
            category_name = category_names[i]
            mano_layer = mano_layer_map[category_name]
            pose = np.array(data['pose'][i])
            shape = np.array(data['shape'][i])
            trans = np.array(data['trans'][i])
            _, joints_fk = mano_layer.forward(
                torch.tensor(pose.reshape(1, -1), dtype=torch.float32),
                torch.tensor(shape.reshape(1,-1), dtype=torch.float32),
                torch.tensor(trans.reshape(1,-1), dtype=torch.float32)
            )
            joints_fk = 0.001 * joints_fk[0].numpy()    
            if 'D' in data['cam_param'] and 'Xi' in data['cam_param'] and 'K' in data['cam_param']:
              K = np.array(data['cam_param']['K']).astype(np.float32)
              Xi = float(data['cam_param']['Xi'])
              D = np.array(data['cam_param']['D']).astype(np.float32)
              
              joints_fk = joints_fk.reshape(joints_fk.shape[0],1,3).astype(np.float32)
              uv,_ = cv2.omnidir.projectPoints(joints_fk, np.zeros(3), np.zeros(3), K, Xi, D)
              points_2d = uv[:,0,:].tolist()
              points_2d_list.append(points_2d)
            elif 'K' in data['cam_param']:
              points_2d = np.matmul(K, np.transpose(joints_fk))
              points_2d = np.transpose(points_2d).tolist()
              points_2d_list.append(points_2d)
            
          #update elem
          data.update({
            'joints2d': points_2d_list
          })      
        except:
          logger.error('Couldnt process by mano')
    
  return data_group


class PrevApiHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    user = self.get_current_user()
    if user is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    step = self.get_argument('step', None)
    data = self.get_argument("data", None)
    if step is None or data is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'request parameter wrong')
      return

    user_name = user['full_name']
    if user_name not in self.db['user_record']:
      self.db['user_record'] = []
    step = int(step)
    if step < 0 or step >= len(self.db['user_record'][user_name]):
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'request parameter wrong')
      return

    # 保存当前修改到内存
    data = json.loads(data)
    entry_id = self.db['user_record'][user_name][step]
    self.db['data'][entry_id] = {'value': data, 'status': True, 'time': time.time()}

    # 保存当前修改到文件
    # 使用 offset
    state = self.db['state']
    offset = self.db['dataset'][state]['offset']
    if not os.path.exists(os.path.join(self.dump_folder, state)):
      os.makedirs(os.path.join(self.dump_folder, state))

    try:
      # 发现数据id
      data_id = None
      for item in data:
        if item['title'] == 'ID':
          data_id = str(item['data'])
      if data_id is None:
        data_id = str(entry_id+offset)

      with open(os.path.join(self.dump_folder, state, '%s.json'%data_id), "w") as file_obj:
        json.dump(data, file_obj)
    except Exception as e:
      print('str(Exception):\t', str(Exception))
      print('str(e):\t\t', str(e))
      print('repr(e):\t', repr(e))
      print('e.message:\t', e.message)

    # 获得当前用户上一步数据
    if step == 0:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'dont have pre step')
      return

    pre_step = step - 1
    pre_entry_id = self.db['user_record'][user_name][pre_step]
    
    try:
      response_content = {
        'value': update_vis_elem(self.db['data'][pre_entry_id]['value']),
        'step': pre_step,
        'tags': self.settings.get('tags', []),
        'operators': [],
        'state': self.db['state'],
        'samples_num': self.db['dataset'][self.db['state']]['samples_num'],
        'samples_num_checked': self.db['dataset'][self.db['state']]['samples_num_checked'],
      }
    except Exception:
      print(traceback.format_exc())
      self.response(RESPONSE_STATUS_CODE.INTERNAL_SERVER_ERROR)      
      return
    
    self.response(RESPONSE_STATUS_CODE.SUCCESS, content=response_content)


class NextApiHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    user = self.get_current_user()
    if user is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    step = self.get_argument("step", None)
    data = self.get_argument("data", None)
    if step is None or data is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'request parameter wrong')
      return

    user_name = user['full_name']
    if user_name not in self.db['user_record']:
      self.db['user_record'] = []
    step = int(step)
    if step < 0 or step >= len(self.db['user_record'][user_name]):
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'request parameter wrong')
      return

    # 保存当前修改到内存
    data = json.loads(data)
    entry_id = self.db['user_record'][user_name][step]
    if not self.db['data'][entry_id]['status']:
      self.db['dataset'][self.db['state']]['samples_num_checked'] += 1

    self.db['data'][entry_id] = {'value': data, 'status': True, 'time': time.time()}

    # 保存当前修改到文件
    # 使用 offset
    state = self.db['state']
    offset = self.db['dataset'][state]['offset']
    if not os.path.exists(os.path.join(self.dump_folder, state)):
      os.makedirs(os.path.join(self.dump_folder, state))

    try:
      # 发现数据id
      data_id = None
      for item in data:
        assert('id' in item or 'ID' in item)
        if 'id' in item:
          data_id = str(item['id'])
        if 'ID' in item:
          data_id = str(item['ID'])

      if data_id is None:
        data_id = str(entry_id + offset)

      with open(os.path.join(self.dump_folder, state, '%s.json' % data_id), "w") as fp:
        json.dump(data, fp)
    except Exception as e:
      print('str(Exception):\t', str(Exception))
      print('str(e):\t\t', str(e))
      print('repr(e):\t', repr(e))
      print('e.message:\t', e.message)

    # 获得用户下一步数据
    if step < len(self.db['user_record'][user_name]) - 1:
      next_step = step + 1
      next_entry_id = self.db['user_record'][user_name][next_step]
      try:
        response_content = {
          'value': update_vis_elem(self.db['data'][next_entry_id]['value']),
          'step': next_step,
          'tags': self.settings.get('tags', []),
          'operators': [],
          'state': self.db['state'],
          'samples_num': self.db['dataset'][self.db['state']]['samples_num'],
          'samples_num_checked': self.db['dataset'][self.db['state']]['samples_num_checked'],
        }
      except Exception:
        print(traceback.format_exc())
        self.response(RESPONSE_STATUS_CODE.INTERNAL_SERVER_ERROR)
        return
      
      self.response(RESPONSE_STATUS_CODE.SUCCESS, content=response_content)
      return

    # 发下下一个还没有进行审核的样本
    next_entry_id = -1
    for id in range(len(self.db['data'])):
      if 'status' not in self.db['data'][id] or not self.db['data'][id]['status']:
        next_entry_id = id
        break

    if next_entry_id == -1:
      try:
        self.response(RESPONSE_STATUS_CODE.SUCCESS, content={
          'value': update_vis_elem(self.db['data'][self.db['user_record'][user_name][-1]]['value']),
          'step': len(self.db['user_record'][user_name]) - 1,
          'tags': self.settings.get('tags', []),
          'operators': [],
          'state': self.db['state'],
          'samples_num': self.db['dataset'][self.db['state']]['samples_num'],
          'samples_num_checked': self.db['dataset'][self.db['state']]['samples_num_checked'],
        })
      except Exception:
        print(traceback.format_exc())
        self.response(RESPONSE_STATUS_CODE.INTERNAL_SERVER_ERROR)
      return

    # 为当前用户分配下一个审查数据
    self.db['user_record'][user_name].append(next_entry_id)

    #
    try:
      response_content = {
        'value': update_vis_elem(self.db['data'][next_entry_id]['value']),
        'step': len(self.db['user_record'][user_name]) - 1,
        'tags': self.settings.get('tags', []),
        'operators': [],
        'state': self.db['state'],
        'samples_num': self.db['dataset'][self.db['state']]['samples_num'],
        'samples_num_checked': self.db['dataset'][self.db['state']]['samples_num_checked'],
      }
    except Exception:
      print(traceback.format_exc())
      self.response(RESPONSE_STATUS_CODE.INTERNAL_SERVER_ERROR)
      return

    self.response(RESPONSE_STATUS_CODE.SUCCESS, content=response_content)
    return
  
  
class RandomApiHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    user = self.get_current_user()
    if user is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    step = self.get_argument("step", None)
    data = self.get_argument("data", None)
    if step is None or data is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'request parameter wrong')
      return

    user_name = user['full_name']
    if user_name not in self.db['user_record']:
      self.db['user_record'] = []
    step = int(step)
    if step < 0 or step >= len(self.db['user_record'][user_name]):
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'request parameter wrong')
      return

    # 保存当前修改到内存
    data = json.loads(data)
    entry_id = self.db['user_record'][user_name][step]
    if not self.db['data'][entry_id]['status']:
      self.db['dataset'][self.db['state']]['samples_num_checked'] += 1

    self.db['data'][entry_id] = {'value': data, 'status': True, 'time': time.time()}

    # 保存当前修改到文件
    # 使用 offset
    state = self.db['state']
    offset = self.db['dataset'][state]['offset']
    if not os.path.exists(os.path.join(self.dump_folder, state)):
      os.makedirs(os.path.join(self.dump_folder, state))

    try:
      # 发现数据id
      data_id = None
      for item in data:
        assert('id' in item or 'ID' in item)
        if 'id' in item:
          data_id = str(item['id'])
        if 'ID' in item:
          data_id = str(item['ID'])

      if data_id is None:
        data_id = str(entry_id + offset)

      with open(os.path.join(self.dump_folder, state, '%s.json' % data_id), "w") as fp:
        json.dump(data, fp)
    except Exception as e:
      print('str(Exception):\t', str(Exception))
      print('str(e):\t\t', str(e))
      print('repr(e):\t', repr(e))
      print('e.message:\t', e.message)

    ################  随机找寻一个样本  ##################
    accessed_ids =  self.db['user_record'][user_name]
    candidate_ids = [i for i in range(len(self.db['data'])) if i not in accessed_ids]
    if len(candidate_ids) == 0:
      # 已经没有未访问的数据
      candidate_ids = [i for i in range(len(self.db['data']))]
    next_entry_id = int(np.random.choice(candidate_ids))
    self.db['user_record'][user_name].append(next_entry_id)

    #
    try:
      response_content = {
        'value': update_vis_elem(self.db['data'][next_entry_id]['value']),
        'step': len(self.db['user_record'][user_name]) - 1,
        'tags': self.settings.get('tags', []),
        'operators': [],
        'state': self.db['state'],
        'samples_num': self.db['dataset'][self.db['state']]['samples_num'],
        'samples_num_checked': self.db['dataset'][self.db['state']]['samples_num_checked'],
      }
    except Exception:
      print(traceback.format_exc())
      self.response(RESPONSE_STATUS_CODE.INTERNAL_SERVER_ERROR)      
      return 

    self.response(RESPONSE_STATUS_CODE.SUCCESS, content=response_content)
    return


class FileApiHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    # 获得代码目录结构
    root_path = self.data_folder
    files_tree = [{
      'name': ".",
      'type': "folder",
      'size': "",
      'folder': [],
      'path': ''
    }]
    queue = [files_tree[-1]]
    while len(queue) != 0:
      folder = queue.pop(-1)
      folder_path = os.path.join(root_path, folder['path'])
      for f in os.listdir(folder_path):
        if os.path.isdir(os.path.join(folder_path, f)):
          folder['folder'].append({
            'name': f,
            'type': "folder",
            'size': "",
            'folder': [],
            'path': '%s/%s' % (folder['path'], f) if folder['path'] != '' else f,
          })

          queue.append(folder['folder'][-1])
        else:
          fsize = os.path.getsize(os.path.join(folder_path, f))
          fsize = fsize / 1024.0  # KB
          folder['folder'].append({
            'name': f,
            'type': "file",
            'size': "%0.2f KB" % round(fsize, 2),
            'path': '%s/%s' % (folder['path'], f) if folder['path'] != '' else f,
          })

    response = {
      'files_tree': files_tree
    }
    self.response(RESPONSE_STATUS_CODE.SUCCESS, content=response)


class DownloadApiHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    if not os.path.exists(os.path.join(self.dump_folder, self.db['state'])):
      self.finish()
      return
    
    package_data = {}
    for file_name in os.listdir(os.path.join(self.dump_folder, self.db['state'])):
      if not file_name.endswith('json'):
        continue
      
      data_id = file_name.split('.')[0]
      with open(os.path.join(self.dump_folder, self.db['state'], file_name), 'r') as fp:
        package_data[data_id] = json.load(fp)
              
    now_time = time.strftime("%Y-%m-%dx%H:%M:%S", time.localtime(time.time()))
    download_file = f'{now_time}.json'
    download_path = os.path.join(self.dump_folder, download_file)
    with open(download_path, 'w') as fp:
      json.dump(package_data, fp)

    # Content-Type
    self.set_header('Content-Type', 'application/octet-stream')
    self.set_header('Content-Disposition', 'attachment; filename=' + download_file)

    buffer_size = 64 * 1024
    with open(download_path, 'rb') as fp:
      content = fp.read()
      content_size = len(content)
      buffer_segments = content_size // buffer_size
      for buffer_seg_i in range(buffer_segments):
        buffer_data = content[buffer_seg_i * buffer_size: (buffer_seg_i + 1) * buffer_size]
        yield self.write(buffer_data)

      yield self.write(content[buffer_segments * buffer_size:])

    self.finish()


class OperatorApiHandler(BaseHandler):
  def post(self):
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class ConfigApiHandler(BaseHandler):
  def post(self):
    offset_config = self.get_argument('offset_config', None)
    if offset_config is not None:
      offset_config = json.loads(offset_config)
      dataset_flag = offset_config['dataset_flag']
      dataset_offset = offset_config['dataset_offset']
      if 'dataset' not in self.db:
        self.db['dataset'] = {}
      if dataset_flag not in self.db['dataset']:
        self.db['dataset'][dataset_flag] = {}

      self.db['dataset'][dataset_flag]['offset'] = dataset_offset

    profile_config = self.get_argument('profile_config', None)
    if profile_config is not None:
      profile_config = json.loads(profile_config)
      dataset_flag = profile_config['dataset_flag']
      samples_num = profile_config['samples_num']
      samples_num_checked = profile_config['samples_num_checked']

      if 'dataset' not in self.db:
        self.db['dataset'] = {}
      if dataset_flag not in self.db['dataset']:
        self.db['dataset'][dataset_flag] = {}

      self.db['dataset'][dataset_flag]['samples_num'] = samples_num
      self.db['dataset'][dataset_flag]['samples_num_checked'] = samples_num_checked

    state = self.get_argument('state', None)
    if state is not None:
      # train,val or test
      self.db['state'] = state

class LoginHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    user_name = self.get_argument('user_name', None)
    user_password = self.get_argument('user_password', None)
    if user_name is None or user_password is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    white_users = self.db['white_users']
    if white_users is None:
      # 无需登录
      self.response(RESPONSE_STATUS_CODE.SUCCESS)
      return

    if user_name not in white_users:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    if user_password != self.db['white_users'][user_name]['password']:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    cookie_id = str(uuid.uuid4())
    self.db['users'].update({
      cookie_id: {
        "full_name": user_name,
        'short_name': user_name[0].upper(),
      }
    })
    self.set_login_cookie({'cookie_id': cookie_id})
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class LogoutHandler(BaseHandler):
  @gen.coroutine
  def post(self):
    self.clear_cookie('antgo')


class UserInfoHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    # 获取当前用户名
    user = self.get_current_user()
    if user is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    # 遍历所有样本获得本轮，当前用户的标注记录
    statistic_info = {
    }

    self.response(RESPONSE_STATUS_CODE.SUCCESS, content={
      'user_name': user['full_name'],
      'short_name': user['short_name'],
      'task_name': 'DEFAULT',
      'task_type': 'DEFAULT',
      'project_type': 'BROWSER',
      'statistic_info': statistic_info
    })


class ProjectInfoHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    
    self.response(RESPONSE_STATUS_CODE.SUCCESS, content={
      'project_type': 'BROWSER',
      'project_state': {
        'stage': \
          'finish' if len(self.db['data']) > 0 and \
            self.db['dataset'][self.db['state']]['samples_num_checked'] == len(self.db['data'])  else 'checking',
        'need_input': self.settings.get('need_input', False),
      },      
    })
    return

class PingApiHandler(BaseHandler):
  @gen.coroutine
  def get(self, *args, **kwargs):
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class GracefulExitException(Exception):
  @staticmethod
  def sigterm_handler(signum, frame):
    raise GracefulExitException()


def browser_server_start(browser_dump_dir,
                         tags,
                         server_port,
                         offset_configs,
                         profile_config,
                         sample_folder=None, sample_list=None, sample_meta={}, need_input=False,
                         white_users=None):
  # register sig
  signal.signal(signal.SIGTERM, GracefulExitException.sigterm_handler)

  # local ext module
  global mano_layer_map
  try:
    from manopth.manolayer import ManoLayer
    left_mano_layer = ManoLayer(
      mano_root="extra/models/mano", use_pca=False, flat_hand_mean=False, side='left'
    )   
    right_mano_layer = ManoLayer(
      mano_root="extra/models/mano", use_pca=False, flat_hand_mean=False, side='right'
    )   
    mano_layer_map = {
      'left': left_mano_layer,
      'right': right_mano_layer
    }
  except Exception:
    print(traceback.format_exc())
    logger.warn('Couldnt load mano model.')


  # 0.step define http server port
  define('port', default=server_port, help='run on port')

  try:
    # 静态资源目录
    browser_dir = os.path.join(browser_dump_dir, 'browser')
    static_dir = os.path.join(browser_dir, 'static')
    if not os.path.exists(static_dir):
      os.makedirs(static_dir)

    # 数据数据目录
    browser_dump_dir = os.path.join(browser_dump_dir, 'record')
    if not os.path.exists(browser_dump_dir):
      os.makedirs(browser_dump_dir)

    # 2.step launch show server
    db = {'data': [], 'users': {}, 'dataset': {}, 'user_record': {}}

    if sample_list is not None and sample_folder is not None:
      # 为样本所在目录建立软连接到static下面
      sample_folder = os.path.abspath(sample_folder)  # 将传入的路径转换为绝对路径    
      os.system(f'cd {static_dir}; ln -s {sample_folder} dataset;')
      
      # 将数据信息写如本地数据库
      # 1.step 尝试加载meta信息
      meta_info = sample_meta
      if len(meta_info) == 0:
        meta_info.update({
          'meta': {}
        })
      # 2.step 样本信息写入数据库
      for sample_id, sample in enumerate(sample_list):
        file_name = sample['image_file'] if sample['image_file'] != '' else sample['image_url']
        data_source = ''
        if sample['image_url'] != '':
          data_source = sample['image_url']
        else:
          data_source = f'/static/dataset/{sample["image_file"]}'

        sample_label = ''
        if 'image_label_name' in sample:
          sample_label = sample['image_label_name']
        elif 'image_label' in sample:
          sample_label = sample['image_label']        
        convert_sample = [{
          'type': 'IMAGE',
          'data': data_source,
          'tag': [sample_label],
          'title': file_name,
          'id': sample_id
        }]
        
        #############################  标准信息  ###############################
        # 添加框元素
        if 'bboxes' in sample:
          convert_sample[0].update({
            'bboxes': sample['bboxes']
          })
        # 添加多边形元素
        if 'segments' in sample:
            convert_sample[0].update({
            'segments': sample['segments']
          })          
        # 添加框标签信息
        if 'labels' in sample:
          convert_sample[0].update({
            'labels': sample['labels']
          })
        # 添加样本标签
        if 'image_label' in sample:
            convert_sample[0].update({
            'image_label': sample['image_label']
          })
            
        #############################  扩展信息 1  #############################
        # 添加2d关键点元素
        if 'joints2d' in sample:
          convert_sample[0].update({
            'joints2d': sample['joints2d'],
            'skeleton': meta_info['meta']['skeleton'] if 'skeleton' in meta_info['meta'] else []
          })
        
        #############################  扩展信息 2  #############################
        # 添加3d关键点元素（需要相机参数）
        if 'joints3d' in sample and \
          'cam_param' in sample and len(sample['cam_param']) > 0:
          convert_sample.append({
            'type': 'IMAGE',
            'data': data_source,
            'tag': [sample_label],
            'title': f'3D-POINTS-({file_name})',
            'id': sample_id
          })
          convert_sample[-1].update({
            'joints3d': sample['joints3d'],
            'skeleton': meta_info['meta']['skeleton'] if 'skeleton' in meta_info['meta'] else [],
            'cam_param': sample['cam_param']
          })
        
        #############################  扩展信息 3  #############################    
        # 添加3d关键点元素（mano）(需要相机参数)
        if 'pose' in sample and 'trans' in sample and 'shape' in sample and \
          'cam_param' in sample and len(sample['cam_param']) > 0 and \
          'model' in meta_info['meta']:
          
          pose_shape_model = meta_info['meta']['model']
          convert_sample.append({
            'type': 'IMAGE',
            'data': data_source,
            'tag': [sample_label],
            'title': f'{pose_shape_model}-({file_name})',
            'id': sample_id
          })
          convert_sample[-1].update({
            'joints3d': sample['joints3d'],
            'skeleton': meta_info['meta']['skeleton'] if 'skeleton' in meta_info['meta'] else [],
            'labels': sample['labels'] if 'labels' in sample else [],
            'category': meta_info['meta']['category'] if 'category' in meta_info['meta'] else {},
            'pose': sample['pose'],
            'shape': sample['shape'],
            'trans': sample['trans'],
            'cam_param': sample['cam_param'],
            'model': pose_shape_model
          })          
                
        db['data'].append({
          'value': convert_sample,
          'status': False,
          'time': time.time()
        })

    # 设置白盒用户
    db['white_users'] = white_users

    for offset_config in offset_configs:
      db['dataset'][offset_config['dataset_flag']] = {
        'offset': offset_config['dataset_offset']
      }

    db['dataset'][profile_config['dataset_flag']]['samples_num'] = \
      profile_config['samples_num']
    db['dataset'][profile_config['dataset_flag']]['samples_num_checked'] = \
      profile_config['samples_num_checked']
    db['state'] = profile_config['dataset_flag']

    # cookie
    cookie_secret = base64.b64encode(uuid.uuid4().bytes + uuid.uuid4().bytes)

    settings = {
      'static_path': os.path.join(browser_dir, 'static'),
      'dump_path': browser_dump_dir,
      'port': server_port,
      'tags': tags,
      'need_input': need_input,
      'cookie_secret': cookie_secret,
      'cookie_max_age_days': 30,
      'Content-Security-Policy': "frame-ancestors 'self' {}".format('http://localhost:8080/'),
      'db': db,
    }

    app = tornado.web.Application(handlers=[(r"/antgo/api/browser/sample/prev/", PrevApiHandler),
                                            (r"/antgo/api/browser/sample/next/", NextApiHandler),
                                            (r"/antgo/api/browser/sample/random/", RandomApiHandler),
                                            (r'/antgo/api/browser/sample/fresh/', FreshApiHandler),
                                            (r"/antgo/api/browser/sample/entry/", EntryApiHandler),
                                            (r"/antgo/api/browser/operators/", OperatorApiHandler),
                                            (r"/antgo/api/browser/file/", FileApiHandler),
                                            (r"/antgo/api/browser/download/", DownloadApiHandler),
                                            (r"/antgo/api/browser/config/", ConfigApiHandler),
                                            (r"/antgo/api/ping/", PingApiHandler),
                                            (r"/antgo/api/user/login/", LoginHandler),    # 登录，仅支持预先指定用户
                                            (r"/antgo/api/user/logout/", LogoutHandler),  # 退出
                                            (r"/antgo/api/user/info/", UserInfoHandler),  # 获得用户信息
                                            (r"/antgo/api/info/", ProjectInfoHandler),
                                            (r'/(.*)', tornado.web.StaticFileHandler,
                                             {"path": static_dir, "default_filename": "index.html"}),],
                                  **settings)
    http_server = tornado.httpserver.HTTPServer(app)
    http_server.listen(options.port)

    logger.info('browser is providing server on port %d' % server_port)
    tornado.ioloop.IOLoop.instance().start()
    logger.info('browser stop server')
  except GracefulExitException:
    logger.info('browser server exit')
    sys.exit(0)
  except KeyboardInterrupt:
    pass


if __name__ == '__main__':
  data_path = ''
  browser_dump_dir = '/Users/jian/Downloads/BB'
  tags = ['A','B','D']
  server_port = 9000
  offset_configs = [{
    'dataset_flag': 'TRAIN',
    'dataset_offset': 0
  }, {
    'dataset_flag': 'VAL',
    'dataset_offset': 0
  }, {
    'dataset_flag': 'TEST',
    'dataset_offset': 0
  }]
  profile_config = {
    'dataset_flag': 'TRAIN',
    'samples_num': 10,
    'samples_num_checked': 0
  }
  white_users = {
    'jian@baidu.com':{
      'password': '112233'
    }
  }
  samples=[]
  for _ in range(10):
    temp = [
      {
        'type': 'IMAGE',
        'data': '/static/data/1.jpeg',
        'width': 1200//4,
        'height': 800//4,
        'tag': ['A'],
        'title': 'MIAO'
      },
      {
        'type': 'STRING',
        'data': 'AABBCC',
        'tag': ['B'],
        'title': 'H'
      }
    ]

    samples.append(copy.deepcopy(temp))

  browser_server_start(
    browser_dump_dir,
    tags,
    server_port,
    offset_configs,
    profile_config,
    white_users=white_users,
    sample_list=samples, sample_folder='')



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from ...builder import DISTILL_LOSSES


def conv1x1(in_channels, out_channels):
	return nn.Conv2d(in_channels, out_channels,
					 kernel_size=1, stride=1,
					 padding=0, bias=False)

'''
Modified from https://github.com/HobbitLong/RepDistiller/blob/master/distiller_zoo/VID.py
'''
@DISTILL_LOSSES.register_module()
class VID(nn.Module):
	'''
	Variational Information Distillation for Knowledge Transfer
	https://zpascal.net/cvpr2019/Ahn_Variational_Information_Distillation_for_Knowledge_Transfer_CVPR_2019_paper.pdf
	'''
	def __init__(self, in_channels, mid_channels, out_channels, init_var, eps=1e-6):
		super(VID, self).__init__()
		self.eps = eps
		self.regressor = nn.Sequential(*[
				conv1x1(in_channels, mid_channels),
				# nn.BatchNorm2d(mid_channels),
				nn.ReLU(),
				conv1x1(mid_channels, mid_channels),
				# nn.BatchNorm2d(mid_channels),
				nn.ReLU(),
				conv1x1(mid_channels, out_channels),
			])
		self.alpha = nn.Parameter(
				np.log(np.exp(init_var-eps)-1.0) * torch.ones(out_channels)
			)

		for m in self.modules():
			if isinstance(m, nn.Conv2d):
				nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
				if m.bias is not None:
					nn.init.constant_(m.bias, 0)
			# elif isinstance(m, nn.BatchNorm2d):
			# 	nn.init.constant_(m.weight, 1)
			# 	nn.init.constant_(m.bias, 0)

	def forward(self, fm_s, fm_t):
		pred_mean = self.regressor(fm_s)
		pred_var  = torch.log(1.0+torch.exp(self.alpha)) + self.eps
		pred_var  = pred_var.view(1, -1, 1, 1)
		neg_log_prob = 0.5 * (torch.log(pred_var) + (pred_mean-fm_t)**2 / pred_var)
		loss = torch.mean(neg_log_prob)

		return loss


import torch.nn as nn
import torch.nn.functional as F
from antgo.framework.helper.models.builder import BACKBONES
from antgo.framework.helper.cnn import ConvModule,DepthwiseSeparableConvModule


class InverseBottleneck(nn.Module):
    expansion = 1
    exp = 1

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None
    ):
        super(InverseBottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError("InverseBottleneck only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in InverseBottleneck")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = nn.Conv2d(inplanes, planes * self.exp, 1, 1, 0)
        self.bn1 = norm_layer(planes * self.exp)
        self.relu1 = nn.ReLU(inplace=True)        
        self.conv2 = nn.Conv2d(
                    planes * self.exp,
                    planes * self.exp,
                    kernel_size=3,
                    stride=stride,
                    padding=1,
                    groups=planes * self.exp,
                    bias=False,
                    dilation=1,
                )
        self.bn2 = norm_layer(planes * self.exp)
        self.relu2 = nn.ReLU(inplace=True)
        self.conv3 = nn.Conv2d(planes * self.exp, planes, 1, 1, 0)
        self.bn3 = norm_layer(planes)
        self.relu3 = nn.ReLU(inplace=True)

        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu1(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu2(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu3(out)

        return out


@BACKBONES.register_module()
class DDRKetNetF(nn.Module):
    def __init__(self, architecture='resnet34', in_channels=1):
        super(DDRKetNetF, self).__init__()
        self.norm_layer = nn.BatchNorm2d
        assert architecture in ["resnet18", "resnet34", "resnet50", "resnet101", "resnet152"]
        layers = {
            "resnet34": [1, 1, 1, 2, 2, 1, 4, 4],
        }
        self.inplanes = 32
        if architecture == "resnet18" or architecture == "resnet34":
            self.block = InverseBottleneck
        else:
            self.block = InverseBottleneck
        self.layers = layers[architecture]

        self.conv1 = nn.Conv2d(in_channels, 32, kernel_size=5, stride=2, padding=2, bias=False)
        self.bn1 = self.norm_layer(32, eps=1e-5, momentum=0.1, affine=True)
        self.relu1 = nn.ReLU(inplace=True)

        self.block.exp = 3
        self.layer1 = self.make_layer(self.block, 32, self.layers[0], stride=1)
        self.layer2 = self.make_layer(self.block, 32, self.layers[1], stride=2)
        self.block.exp = 6
        self.layer3 = self.make_layer(self.block, 32, self.layers[2], stride=1)
        self.layer4 = self.make_layer(self.block, 64, self.layers[3], stride=2)
        self.block.exp = 4
        self.layer5 = self.make_layer(self.block, 96, self.layers[4], stride=1)
        self.layer6 = self.make_layer(self.block, 128, self.layers[5], stride=2)
        self.block.exp = 6
        self.layer7 = self.make_layer(self.block, 128, self.layers[6], stride=1)
        self.layer8 = self.make_layer(self.block, 160, self.layers[7], stride=2)

        # 1/4 ,  1/8, 1/16, 1/32
        # (dilation(k-1)+1-stride)/2
        self.down4_to_16 = \
                ConvModule(
                    32, 
                    128,
                    kernel_size=3,
                    stride=4, 
                    padding=(int)((4*(3-1)+1-4)//2+1), 
                    dilation=4, 
                    act_cfg=None, 
                    norm_cfg=dict(type='BN'))
        self.down4_to_32 = \
                ConvModule(
                    32, 
                    160,
                    kernel_size=3,
                    stride=8, 
                    padding=(int)((6*(3-1)+1-8)//2+1), 
                    dilation=6, 
                    act_cfg=None, 
                    norm_cfg=dict(type='BN'))             

        self.compression8_to_4 = \
                ConvModule(
                    96, 
                    32, 
                    kernel_size=1, 
                    act_cfg=None, 
                    norm_cfg=dict(type='BN'))
        self.compression16_to_4 = \
                ConvModule(
                    128, 
                    32, 
                    kernel_size=1, 
                    act_cfg=None, 
                    norm_cfg=dict(type='BN'))

        self.compression32_to_4 = \
                ConvModule(
                    160, 
                    32, 
                    kernel_size=1, 
                    act_cfg=None, 
                    norm_cfg=dict(type='BN'))
    
        self.layer4_1 = nn.Sequential(
            ConvModule(
                32,
                32,
                kernel_size=3,
                padding=1,
                act_cfg=dict(type='ReLU'),
                norm_cfg=dict(type='BN'),
            )
        )
        self.layer4_2 = nn.Sequential(
            ConvModule(
                32,
                32,
                kernel_size=3,
                padding=1,
                act_cfg=dict(type='ReLU'),
                norm_cfg=dict(type='BN'),
            )
        )      
        self.layer4_3 = nn.Sequential(
            ConvModule(
                32,
                32,
                kernel_size=3,
                padding=1,
                act_cfg=dict(type='ReLU'),
                norm_cfg=dict(type='BN'),
            )
        )      
        self.layer4_4 = nn.Sequential(
            ConvModule(
                32,
                32,
                kernel_size=3,
                padding=1,
                act_cfg=dict(type='ReLU'),
                norm_cfg=dict(type='BN'),
            )
        )       

    def forward(self, x):
        x = self.relu1(self.bn1(self.conv1(x)))  # 32 * h/2 * w/2
        x = self.layer1(x)
        x32 = self.layer2(x)    # 32 * h/4 * w/4
        x32 = self.layer3(x32)  # 64 * h/4 * w/4

        out_layer_1 = self.layer4_1(x32)

        x16 = self.layer4(x32)  # 64 * h/8 * w/8
        x16 = self.layer5(x16)  # 96 * h/8 * w/8

        # 
        out_layer_2 = \
            self.layer4_2(out_layer_1) + F.interpolate(
                                    self.compression8_to_4(x16),
                                    scale_factor=2,
                                    mode='bilinear')
        
        x8 = self.layer6(x16)   # 128 * h/16 * w/16
        x8 = self.layer7(x8)    # 128 * h/16 * w/16

        down_x = F.relu(x8 + self.down4_to_16(out_layer_2))
        out_layer_3 = \
            self.layer4_3(out_layer_2) + F.interpolate(
                                self.compression16_to_4(x8),
                                scale_factor=4,
                                mode='bilinear')
        x8 = down_x
        x4 = self.layer8(x8)    # 160 * h/32 * w/32
        x4 = F.relu(x4 + self.down4_to_32(out_layer_3))

        out_layer = \
            self.layer4_4(out_layer_3) + F.interpolate(
                self.compression32_to_4(x4), 
                scale_factor=8, 
                mode='bilinear')
        return [out_layer]

    def stages(self):
        return [self.layer1, self.layer2, self.layer3, self.layer4]

    def make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion, kernel_size=1, stride=stride, bias=False),
                self.norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample, norm_layer=self.norm_layer))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, norm_layer=self.norm_layer))

        return nn.Sequential(*layers)



import os.path as osp
import warnings
from antgo.framework.helper.fileio import FileClient
from ..dist_utils import allreduce_params, master_only
from .hook import HOOKS, Hook


@HOOKS.register_module()
class CheckpointHook(Hook):
    """Save checkpoints periodically.

    Args:
        interval (int): The saving period. If ``by_epoch=True``, interval
            indicates epochs, otherwise it indicates iterations.
            Default: -1, which means "never".
        by_epoch (bool): Saving checkpoints by epoch or by iteration.
            Default: True.
        save_optimizer (bool): Whether to save optimizer state_dict in the
            checkpoint. It is usually used for resuming experiments.
            Default: True.
        out_dir (str, optional): The root directory to save checkpoints. If not
            specified, ``runner.work_dir`` will be used by default. If
            specified, the ``out_dir`` will be the concatenation of ``out_dir``
            and the last level directory of ``runner.work_dir``.
            `Changed in version 1.3.16.`
        max_keep_ckpts (int, optional): The maximum checkpoints to keep.
            In some cases we want only the latest few checkpoints and would
            like to delete old ones to save the disk space.
            Default: -1, which means unlimited.
        save_last (bool, optional): Whether to force the last checkpoint to be
            saved regardless of interval. Default: True.
        sync_buffer (bool, optional): Whether to synchronize buffers in
            different gpus. Default: False.
        file_client_args (dict, optional): Arguments to instantiate a
            FileClient. See :class:`mmcv.fileio.FileClient` for details.
            Default: None.
            `New in version 1.3.16.`

    .. warning::
        Before v1.3.16, the ``out_dir`` argument indicates the path where the
        checkpoint is stored. However, since v1.3.16, ``out_dir`` indicates the
        root directory and the final path to save checkpoint is the
        concatenation of ``out_dir`` and the last level directory of
        ``runner.work_dir``. Suppose the value of ``out_dir`` is "/path/of/A"
        and the value of ``runner.work_dir`` is "/path/of/B", then the final
        path will be "/path/of/A/B".
    """

    def __init__(self,
                 interval=-1,
                 by_epoch=True,
                 save_optimizer=True,
                 out_dir=None,
                 max_keep_ckpts=-1,
                 save_last=True,
                 sync_buffer=False,
                 file_client_args=None,
                 **kwargs):
        self.interval = interval
        self.by_epoch = by_epoch
        self.save_optimizer = save_optimizer
        self.out_dir = out_dir
        self.max_keep_ckpts = max_keep_ckpts
        self.save_last = save_last
        self.args = kwargs
        self.sync_buffer = sync_buffer
        self.file_client_args = file_client_args

    def before_run(self, runner):
        if not self.out_dir:
            # 在out_dir没有被设置时，继承runner.work_dir的工作目录
            self.out_dir = runner.work_dir

        self.file_client = \
            FileClient.infer_client(
                self.file_client_args, self.out_dir)

        # 添加固定路径格式
        self.out_dir = osp.join(self.out_dir, 'output', 'checkpoint')
        runner.logger.info(
            (f'Checkpoints will be saved to {self.out_dir} by '
                            f'{self.file_client.name}.'))

        # disable the create_symlink option because some file backends do not
        # allow to create a symlink
        self.args['create_symlink'] = False

    def after_train_epoch(self, runner):
        if not self.by_epoch:
            return

        # save checkpoint for following cases:
        # 1. every ``self.interval`` epochs
        # 2. reach the last epoch of training
        if self.every_n_epochs(
                runner, self.interval) or (self.save_last
                                           and self.is_last_epoch(runner)):
            runner.logger.info(
                f'Saving checkpoint at {runner.epoch + 1} epochs')
            if self.sync_buffer:
                allreduce_params(runner.model.buffers())
            runner.logger.info("after allreduce_params")
            self._save_checkpoint(runner)
            runner.logger.info("after save checkpoint")

    @master_only
    def _save_checkpoint(self, runner):
        """Save the current checkpoint and delete unwanted checkpoint."""
        runner.save_checkpoint(
            self.out_dir, save_optimizer=self.save_optimizer, **self.args)
        if runner.meta is not None:
            if self.by_epoch:
                cur_ckpt_filename = self.args.get(
                    'filename_tmpl', 'epoch_{}.pth').format(runner.epoch + 1)
            else:
                cur_ckpt_filename = self.args.get(
                    'filename_tmpl', 'iter_{}.pth').format(runner.iter + 1)
            runner.meta.setdefault('hook_msgs', dict())
            runner.meta['hook_msgs']['last_ckpt'] = self.file_client.join_path(
                self.out_dir, cur_ckpt_filename)

        # remove other checkpoints
        if self.max_keep_ckpts > 0:
            if self.by_epoch:
                name = 'epoch_{}.pth'
                current_ckpt = runner.epoch + 1
            else:
                name = 'iter_{}.pth'
                current_ckpt = runner.iter + 1
            redundant_ckpts = range(
                current_ckpt - self.max_keep_ckpts * self.interval, 0,
                -self.interval)
            filename_tmpl = self.args.get('filename_tmpl', name)
            for _step in redundant_ckpts:
                ckpt_path = self.file_client.join_path(
                    self.out_dir, filename_tmpl.format(_step))
                if self.file_client.isfile(ckpt_path):
                    self.file_client.remove(ckpt_path)
                else:
                    break

    def after_train_iter(self, runner):
        if self.by_epoch:
            return

        # save checkpoint for following cases:
        # 1. every ``self.interval`` iterations
        # 2. reach the last iteration of training
        if self.every_n_iters(
                runner, self.interval) or (self.save_last
                                           and self.is_last_iter(runner)):
            runner.logger.info(
                f'Saving checkpoint at {runner.iter + 1} iterations')
            if self.sync_buffer:
                allreduce_params(runner.model.buffers())
            self._save_checkpoint(runner)



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/10 23:06
# @File    : base_execution.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

class BaseExecution:
    """
    Execute an operators
    """
    def __apply__(self, *arg, **kws):
        # Multi inputs.
        if isinstance(self._index[0], tuple):
            args = [getattr(arg[0], x) for x in self._index[0]]
        # Single input.
        else:
            args = [getattr(arg[0], self._index[0])]
        
        return self._op(*args, **kws)

    def __call__(self, *arg, **kws):
        self.__check_init__()
        if bool(self._index):
            res = self.__apply__(*arg, **kws)

            # Multi outputs.
            if isinstance(res, tuple):
                if not isinstance(self._index[1],
                                  tuple) or len(self._index[1]) != len(res):
                    raise IndexError(
                        f'Op has {len(res)} outputs, but {len(self._index[1])} indices are given.'
                    )
                for i, j in zip(self._index[1], res):
                    setattr(arg[0], i, j)
            # Single output.
            else:
                setattr(arg[0], self._index[1], res)

            return arg[0]
        else:
            res = self._op(*arg, **kws)
            return res



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/5 23:38
# @File    : __init__.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.pipeline.hub import *
import os

def run(project, folder='./deploy', **kwargs):
    project_folder = os.path.join(folder, f'{project}_plugin')
    if not os.path.exists(project_folder):
        print(f'Project {project} not exist.')
        return

    os.system(f'cd {project_folder} && bash run.sh')

def package(project, folder='./deploy', **kwargs):
    # project_folder = os.path.join(folder, f'{project}_plugin')
    # if not os.path.exists(project_folder):
    #     print(f'Project {project} not exist.')
    #     return

    # os.system(f'cd {project_folder} && bash package.sh')
    pass


def service(project, folder='./deploy', **kwargs):
    pass


# Copyright (c) OpenMMLab. All rights reserved.


def is_ipu_available():
    try:
        import poptorch
        return poptorch.ipuHardwareIsAvailable()
    except ImportError:
        return False


IS_IPU_AVAILABLE = is_ipu_available()


def is_mlu_available():
    try:
        import torch
        return (hasattr(torch, 'is_mlu_available')
                and torch.is_mlu_available())
    except Exception:
        return False


IS_MLU_AVAILABLE = is_mlu_available()



import math
import warnings

import torch
import torch.nn as nn

from antgo.framework.helper.models.detectors.core.bbox import bbox_overlaps
from antgo.framework.helper.models.builder import LOSSES
from .utils import weighted_loss


@weighted_loss
def iou_loss(pred, target, linear=False, mode='log', eps=1e-6):
    """IoU loss.

    Computing the IoU loss between a set of predicted bboxes and target bboxes.
    The loss is calculated as negative log of IoU.

    Args:
        pred (torch.Tensor): Predicted bboxes of format (x1, y1, x2, y2),
            shape (n, 4).
        target (torch.Tensor): Corresponding gt bboxes, shape (n, 4).
        linear (bool, optional): If True, use linear scale of loss instead of
            log scale. Default: False.
        mode (str): Loss scaling mode, including "linear", "square", and "log".
            Default: 'log'
        eps (float): Eps to avoid log(0).

    Return:
        torch.Tensor: Loss tensor.
    """
    assert mode in ['linear', 'square', 'log']
    if linear:
        mode = 'linear'
        warnings.warn('DeprecationWarning: Setting "linear=True" in '
                      'iou_loss is deprecated, please use "mode=`linear`" '
                      'instead.')
    ious = bbox_overlaps(pred, target, is_aligned=True).clamp(min=eps)
    if mode == 'linear':
        loss = 1 - ious
    elif mode == 'square':
        loss = 1 - ious**2
    elif mode == 'log':
        loss = -ious.log()
    else:
        raise NotImplementedError
    return loss


@weighted_loss
def bounded_iou_loss(pred, target, beta=0.2, eps=1e-3):
    """BIoULoss.

    This is an implementation of paper
    `Improving Object Localization with Fitness NMS and Bounded IoU Loss.
    <https://arxiv.org/abs/1711.00164>`_.

    Args:
        pred (torch.Tensor): Predicted bboxes.
        target (torch.Tensor): Target bboxes.
        beta (float): beta parameter in smoothl1.
        eps (float): eps to avoid NaN.
    """
    pred_ctrx = (pred[:, 0] + pred[:, 2]) * 0.5
    pred_ctry = (pred[:, 1] + pred[:, 3]) * 0.5
    pred_w = pred[:, 2] - pred[:, 0]
    pred_h = pred[:, 3] - pred[:, 1]
    with torch.no_grad():
        target_ctrx = (target[:, 0] + target[:, 2]) * 0.5
        target_ctry = (target[:, 1] + target[:, 3]) * 0.5
        target_w = target[:, 2] - target[:, 0]
        target_h = target[:, 3] - target[:, 1]

    dx = target_ctrx - pred_ctrx
    dy = target_ctry - pred_ctry

    loss_dx = 1 - torch.max(
        (target_w - 2 * dx.abs()) /
        (target_w + 2 * dx.abs() + eps), torch.zeros_like(dx))
    loss_dy = 1 - torch.max(
        (target_h - 2 * dy.abs()) /
        (target_h + 2 * dy.abs() + eps), torch.zeros_like(dy))
    loss_dw = 1 - torch.min(target_w / (pred_w + eps), pred_w /
                            (target_w + eps))
    loss_dh = 1 - torch.min(target_h / (pred_h + eps), pred_h /
                            (target_h + eps))
    # view(..., -1) does not work for empty tensor
    loss_comb = torch.stack([loss_dx, loss_dy, loss_dw, loss_dh],
                            dim=-1).flatten(1)

    loss = torch.where(loss_comb < beta, 0.5 * loss_comb * loss_comb / beta,
                       loss_comb - 0.5 * beta)
    return loss


@weighted_loss
def giou_loss(pred, target, eps=1e-7):
    r"""`Generalized Intersection over Union: A Metric and A Loss for Bounding
    Box Regression <https://arxiv.org/abs/1902.09630>`_.

    Args:
        pred (torch.Tensor): Predicted bboxes of format (x1, y1, x2, y2),
            shape (n, 4).
        target (torch.Tensor): Corresponding gt bboxes, shape (n, 4).
        eps (float): Eps to avoid log(0).

    Return:
        Tensor: Loss tensor.
    """
    gious = bbox_overlaps(pred, target, mode='giou', is_aligned=True, eps=eps)
    loss = 1 - gious
    return loss


@weighted_loss
def diou_loss(pred, target, eps=1e-7):
    r"""`Implementation of Distance-IoU Loss: Faster and Better
    Learning for Bounding Box Regression, https://arxiv.org/abs/1911.08287`_.

    Code is modified from https://github.com/Zzh-tju/DIoU.

    Args:
        pred (Tensor): Predicted bboxes of format (x1, y1, x2, y2),
            shape (n, 4).
        target (Tensor): Corresponding gt bboxes, shape (n, 4).
        eps (float): Eps to avoid log(0).
    Return:
        Tensor: Loss tensor.
    """
    # overlap
    lt = torch.max(pred[:, :2], target[:, :2])
    rb = torch.min(pred[:, 2:], target[:, 2:])
    wh = (rb - lt).clamp(min=0)
    overlap = wh[:, 0] * wh[:, 1]

    # union
    ap = (pred[:, 2] - pred[:, 0]) * (pred[:, 3] - pred[:, 1])
    ag = (target[:, 2] - target[:, 0]) * (target[:, 3] - target[:, 1])
    union = ap + ag - overlap + eps

    # IoU
    ious = overlap / union

    # enclose area
    enclose_x1y1 = torch.min(pred[:, :2], target[:, :2])
    enclose_x2y2 = torch.max(pred[:, 2:], target[:, 2:])
    enclose_wh = (enclose_x2y2 - enclose_x1y1).clamp(min=0)

    cw = enclose_wh[:, 0]
    ch = enclose_wh[:, 1]

    c2 = cw**2 + ch**2 + eps

    b1_x1, b1_y1 = pred[:, 0], pred[:, 1]
    b1_x2, b1_y2 = pred[:, 2], pred[:, 3]
    b2_x1, b2_y1 = target[:, 0], target[:, 1]
    b2_x2, b2_y2 = target[:, 2], target[:, 3]

    left = ((b2_x1 + b2_x2) - (b1_x1 + b1_x2))**2 / 4
    right = ((b2_y1 + b2_y2) - (b1_y1 + b1_y2))**2 / 4
    rho2 = left + right

    # DIoU
    dious = ious - rho2 / c2
    loss = 1 - dious
    return loss


@weighted_loss
def ciou_loss(pred, target, eps=1e-7):
    r"""`Implementation of paper `Enhancing Geometric Factors into
    Model Learning and Inference for Object Detection and Instance
    Segmentation <https://arxiv.org/abs/2005.03572>`_.

    Code is modified from https://github.com/Zzh-tju/CIoU.

    Args:
        pred (Tensor): Predicted bboxes of format (x1, y1, x2, y2),
            shape (n, 4).
        target (Tensor): Corresponding gt bboxes, shape (n, 4).
        eps (float): Eps to avoid log(0).
    Return:
        Tensor: Loss tensor.
    """
    # overlap
    lt = torch.max(pred[:, :2], target[:, :2])
    rb = torch.min(pred[:, 2:], target[:, 2:])
    wh = (rb - lt).clamp(min=0)
    overlap = wh[:, 0] * wh[:, 1]

    # union
    ap = (pred[:, 2] - pred[:, 0]) * (pred[:, 3] - pred[:, 1])
    ag = (target[:, 2] - target[:, 0]) * (target[:, 3] - target[:, 1])
    union = ap + ag - overlap + eps

    # IoU
    ious = overlap / union

    # enclose area
    enclose_x1y1 = torch.min(pred[:, :2], target[:, :2])
    enclose_x2y2 = torch.max(pred[:, 2:], target[:, 2:])
    enclose_wh = (enclose_x2y2 - enclose_x1y1).clamp(min=0)

    cw = enclose_wh[:, 0]
    ch = enclose_wh[:, 1]

    c2 = cw**2 + ch**2 + eps

    b1_x1, b1_y1 = pred[:, 0], pred[:, 1]
    b1_x2, b1_y2 = pred[:, 2], pred[:, 3]
    b2_x1, b2_y1 = target[:, 0], target[:, 1]
    b2_x2, b2_y2 = target[:, 2], target[:, 3]

    w1, h1 = b1_x2 - b1_x1, b1_y2 - b1_y1 + eps
    w2, h2 = b2_x2 - b2_x1, b2_y2 - b2_y1 + eps

    left = ((b2_x1 + b2_x2) - (b1_x1 + b1_x2))**2 / 4
    right = ((b2_y1 + b2_y2) - (b1_y1 + b1_y2))**2 / 4
    rho2 = left + right

    factor = 4 / math.pi**2
    v = factor * torch.pow(torch.atan(w2 / h2) - torch.atan(w1 / h1), 2)

    with torch.no_grad():
        alpha = (ious > 0.5).float() * v / (1 - ious + v)

    # CIoU
    cious = ious - (rho2 / c2 + alpha * v)
    loss = 1 - cious.clamp(min=-1.0, max=1.0)
    return loss


@LOSSES.register_module()
class IoULoss(nn.Module):
    """IoULoss.

    Computing the IoU loss between a set of predicted bboxes and target bboxes.

    Args:
        linear (bool): If True, use linear scale of loss else determined
            by mode. Default: False.
        eps (float): Eps to avoid log(0).
        reduction (str): Options are "none", "mean" and "sum".
        loss_weight (float): Weight of loss.
        mode (str): Loss scaling mode, including "linear", "square", and "log".
            Default: 'log'
    """

    def __init__(self,
                 linear=False,
                 eps=1e-6,
                 reduction='mean',
                 loss_weight=1.0,
                 mode='log'):
        super(IoULoss, self).__init__()
        assert mode in ['linear', 'square', 'log']
        if linear:
            mode = 'linear'
            warnings.warn('DeprecationWarning: Setting "linear=True" in '
                          'IOULoss is deprecated, please use "mode=`linear`" '
                          'instead.')
        self.mode = mode
        self.linear = linear
        self.eps = eps
        self.reduction = reduction
        self.loss_weight = loss_weight

    def forward(self,
                pred,
                target,
                weight=None,
                avg_factor=None,
                reduction_override=None,
                **kwargs):
        """Forward function.

        Args:
            pred (torch.Tensor): The prediction.
            target (torch.Tensor): The learning target of the prediction.
            weight (torch.Tensor, optional): The weight of loss for each
                prediction. Defaults to None.
            avg_factor (int, optional): Average factor that is used to average
                the loss. Defaults to None.
            reduction_override (str, optional): The reduction method used to
                override the original reduction method of the loss.
                Defaults to None. Options are "none", "mean" and "sum".
        """
        assert reduction_override in (None, 'none', 'mean', 'sum')
        reduction = (
            reduction_override if reduction_override else self.reduction)
        if (weight is not None) and (not torch.any(weight > 0)) and (
                reduction != 'none'):
            if pred.dim() == weight.dim() + 1:
                weight = weight.unsqueeze(1)
            return (pred * weight).sum()  # 0
        if weight is not None and weight.dim() > 1:
            # TODO: remove this in the future
            # reduce the weight of shape (n, 4) to (n,) to match the
            # iou_loss of shape (n,)
            assert weight.shape == pred.shape
            weight = weight.mean(-1)
        loss = self.loss_weight * iou_loss(
            pred,
            target,
            weight,
            mode=self.mode,
            eps=self.eps,
            reduction=reduction,
            avg_factor=avg_factor,
            **kwargs)
        return loss


@LOSSES.register_module()
class BoundedIoULoss(nn.Module):

    def __init__(self, beta=0.2, eps=1e-3, reduction='mean', loss_weight=1.0):
        super(BoundedIoULoss, self).__init__()
        self.beta = beta
        self.eps = eps
        self.reduction = reduction
        self.loss_weight = loss_weight

    def forward(self,
                pred,
                target,
                weight=None,
                avg_factor=None,
                reduction_override=None,
                **kwargs):
        if weight is not None and not torch.any(weight > 0):
            if pred.dim() == weight.dim() + 1:
                weight = weight.unsqueeze(1)
            return (pred * weight).sum()  # 0
        assert reduction_override in (None, 'none', 'mean', 'sum')
        reduction = (
            reduction_override if reduction_override else self.reduction)
        loss = self.loss_weight * bounded_iou_loss(
            pred,
            target,
            weight,
            beta=self.beta,
            eps=self.eps,
            reduction=reduction,
            avg_factor=avg_factor,
            **kwargs)
        return loss


@LOSSES.register_module()
class GIoULoss(nn.Module):

    def __init__(self, eps=1e-6, reduction='mean', loss_weight=1.0):
        super(GIoULoss, self).__init__()
        self.eps = eps
        self.reduction = reduction
        self.loss_weight = loss_weight

    def forward(self,
                pred,
                target,
                weight=None,
                avg_factor=None,
                reduction_override=None,
                **kwargs):
        if weight is not None and not torch.any(weight > 0):
            if pred.dim() == weight.dim() + 1:
                weight = weight.unsqueeze(1)
            return (pred * weight).sum()  # 0
        assert reduction_override in (None, 'none', 'mean', 'sum')
        reduction = (
            reduction_override if reduction_override else self.reduction)
        if weight is not None and weight.dim() > 1:
            # TODO: remove this in the future
            # reduce the weight of shape (n, 4) to (n,) to match the
            # giou_loss of shape (n,)
            assert weight.shape == pred.shape
            weight = weight.mean(-1)
        loss = self.loss_weight * giou_loss(
            pred,
            target,
            weight,
            eps=self.eps,
            reduction=reduction,
            avg_factor=avg_factor,
            **kwargs)
        return loss


@LOSSES.register_module()
class DIoULoss(nn.Module):

    def __init__(self, eps=1e-6, reduction='mean', loss_weight=1.0):
        super(DIoULoss, self).__init__()
        self.eps = eps
        self.reduction = reduction
        self.loss_weight = loss_weight

    def forward(self,
                pred,
                target,
                weight=None,
                avg_factor=None,
                reduction_override=None,
                **kwargs):
        if weight is not None and not torch.any(weight > 0):
            if pred.dim() == weight.dim() + 1:
                weight = weight.unsqueeze(1)
            return (pred * weight).sum()  # 0
        assert reduction_override in (None, 'none', 'mean', 'sum')
        reduction = (
            reduction_override if reduction_override else self.reduction)
        if weight is not None and weight.dim() > 1:
            # TODO: remove this in the future
            # reduce the weight of shape (n, 4) to (n,) to match the
            # giou_loss of shape (n,)
            assert weight.shape == pred.shape
            weight = weight.mean(-1)
        loss = self.loss_weight * diou_loss(
            pred,
            target,
            weight,
            eps=self.eps,
            reduction=reduction,
            avg_factor=avg_factor,
            **kwargs)
        return loss


@LOSSES.register_module()
class CIoULoss(nn.Module):

    def __init__(self, eps=1e-6, reduction='mean', loss_weight=1.0):
        super(CIoULoss, self).__init__()
        self.eps = eps
        self.reduction = reduction
        self.loss_weight = loss_weight

    def forward(self,
                pred,
                target,
                weight=None,
                avg_factor=None,
                reduction_override=None,
                **kwargs):
        if weight is not None and not torch.any(weight > 0):
            if pred.dim() == weight.dim() + 1:
                weight = weight.unsqueeze(1)
            return (pred * weight).sum()  # 0
        assert reduction_override in (None, 'none', 'mean', 'sum')
        reduction = (
            reduction_override if reduction_override else self.reduction)
        if weight is not None and weight.dim() > 1:
            # TODO: remove this in the future
            # reduce the weight of shape (n, 4) to (n,) to match the
            # giou_loss of shape (n,)
            assert weight.shape == pred.shape
            weight = weight.mean(-1)
        loss = self.loss_weight * ciou_loss(
            pred,
            target,
            weight,
            eps=self.eps,
            reduction=reduction,
            avg_factor=avg_factor,
            **kwargs)
        return loss



# -*- coding: UTF-8 -*-
# Time: 10/10/17
# File: utils.py
# Author: jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import importlib
import os
import sys
from antgo.measures import *
from antgo.dataflow.dataset import *


def main_context(main_file, source_paths):
  # filter .py
  # 1.step load user custom module
  key_model = main_file
  dot_pos = key_model.rfind(".")
  if dot_pos != -1:
    key_model = key_model[0:dot_pos]

  key_model = os.path.normpath(key_model)
  key_model = key_model.replace('/', '.')
  sys.path.append(source_paths)
  module = importlib.import_module(key_model)

  for k, v in module.__dict__.items():
    # 2.step check user custom measure method
    if '_BASE_MEASURE' in dir(v) and k != 'AntMeasure':
      AntMeasuresFactory.add_custom_measure(v)

    # 3.step check user custom dataset parse
    if '_BASE_DATASET' in dir(v) and k != 'Dataset':
      AntDatasetFactory.add_custom_dataset(v)

  return module.get_global_context()


def main_logo():
  logo_str='''

   ,---.      .-._        ,--.--------.     _,---.     _,.---._     
 .--.'  \    /==/ \  .-._/==/,  -   , -\_.='.'-,  \  ,-.' , -  `.   
 \==\-/\ \   |==|, \/ /, |==\.-.  - ,-./==.'-     / /==/_,  ,  - \  
 /==/-|_\ |  |==|-  \|  | `--`\==\- \ /==/ -   .-' |==|   .=.     | 
 \==\,   - \ |==| ,  | -|      \==\_ \|==|_   /_,-.|==|_ : ;=:  - | 
 /==/ -   ,| |==| -   _ |      |==|- ||==|  , \_.' )==| , '='     | 
/==/-  /\ - \|==|  /\ , |      |==|, |\==\-  ,    ( \==\ -    ,_ /  
\==\ _.\=\.-'/==/, | |- |      /==/ -/ /==/ _  ,  /  '.='. -   .'   
 `--`        `--`./  `--`      `--`--` `--`------'     `--`--''     

'''

  print(logo_str)





# -*- coding: UTF-8 -*-
# @Time    : 2022/9/11 23:53
# @File    : storages.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from .entity import EntityView
# pylint: disable=import-outside-toplevel


class WritableTable:
    """
    A wrapper that make arrow table writable.
    """
    def __init__(self, table):
        self._table = table
        self._buffer = {}

    def write_many(self, names, arrays):
        self.prepare()
        if isinstance(arrays, tuple):
            self._buffer = dict(zip(names, arrays))
        else:
            self._buffer[names] = arrays
        return self.seal()

    def write(self, name, offset, value):
        if name not in self._buffer:
            self._buffer[name] = []
        while len(self._buffer[name]) < offset:
            self._buffer[name].append(None)
        self._buffer[name].append(value)

    def prepare(self):
        if not hasattr(self, '_sealed'):
            self._sealed = WritableTable(None)
        return self._sealed

    def seal(self):
        # pylint: disable=protected-access
        from antgo.pipeline.utils.thirdparty.pyarrow import pa
        from antgo.pipeline.types.tensor_array import TensorArray

        names = list(self._buffer)
        arrays = []
        for name in names:
            try:
                arrays.append(pa.array(self._buffer[name]))
            # pylint: disable=bare-except
            except:
                arrays.append(TensorArray.from_numpy(self._buffer[name]))
        new_table = self._table
        for name, arr in zip(names, arrays):
            new_table = new_table.append_column(name, arr)
        self._sealed._table = new_table
        return self._sealed

    def __iter__(self):
        for i in range(self._table.shape[0]):
            yield EntityView(i, self)

    def __len__(self):
        return self._table.num_rows

    def __getattr__(self, name):
        if name in self._table.column_names:
            return self._table.__getitem__(name)
        else:
            return getattr(self._table, name)

    def __getitem__(self, index):
        return self._table.__getitem__(index)

    def __repr__(self) -> str:
        return repr(self._table)


class ChunkedTable:
    """
    Chunked arrow table
    """
    def __init__(self, chunks=None, chunksize=128, stream=False) -> None:
        """
        A chunked pyarrow table.

        Args:
            chunks:
                The list or queue of chunks.
            chunksize (`int`):
                The size of the chunk.
            stream (`bool`):
                If the data is streamed.
        """
        self._chunksize = chunksize
        self._is_stream = stream
        if chunks is not None:
            self._chunks = chunks
        else:
            self._chunks = [] if not stream else None

    @property
    def is_stream(self):
        return self._is_stream

    @property
    def chunksize(self):
        return self._chunksize

    def chunks(self):
        return self._chunks

    def _create_table(self, chunk, head):
        from antgo.pipeline.utils.thirdparty.pyarrow import pa
        from antgo.pipeline.types.tensor_array import TensorArray

        # head = []
        cols = None
        for entity in chunk:
            cols = [[] for _ in head] if cols is None else cols
            for col, name in zip(cols, head):
                col.append(getattr(entity, name))
        arrays = []
        for col in cols:
            try:
                arrays.append(pa.array(col))
            # pylint: disable=bare-except
            except:
                arrays.append(TensorArray.from_numpy(col))

        res = pa.Table.from_arrays(arrays, names=head)

        return res

    def _pack_unstream_chunk(self, data):
        head = []
        res = []
        chunk = []
        # if not self._column_names:
        #     self._column_names = [*data[0].__dict__]
        for element in data:
            if not head:
                head = [*element.__dict__]
            chunk.append(element)
            if len(chunk) >= self._chunksize:
                res.append(WritableTable(self._create_table(chunk, head)))
                chunk = []

        if len(chunk) != 0:
            res.append(WritableTable(self._create_table(chunk, head)))

        return res

    def _pack_stream_chunk(self, data):
        chunk = []
        head = []
        for element in data:
            if not head:
                head = [*element.__dict__]
            chunk.append(element)
            if len(chunk) >= self._chunksize:
                yield WritableTable(self._create_table(chunk, head))
                chunk = []

        if len(chunk) != 0:
            yield WritableTable(self._create_table(chunk, head))


    def feed(self, data):
        if not self._is_stream:
            self._chunks = self._pack_unstream_chunk(data)
        else:
            self._chunks = self._pack_stream_chunk(data)

    def __iter__(self):
        def inner():
            for chunk in self._chunks:
                # yield chunk
                for i in range(len(chunk)):
                    yield EntityView(i, chunk)

        return inner()



import ctypes
import numpy as np
from .common import *

class NumPyTensor(MobulaTensor):
    @property
    def data_ptr(self):
        def p(e):
            return e.ctypes.data_as(ctypes.c_void_p)
        if not self.tensor.flags.c_contiguous:
            c = np.ascontiguousarray(self.tensor)
            return p(c), c
        return p(self.tensor)

    @property
    def ctype(self):
        return NPDTYPE2CTYPE(self.tensor.dtype)

    @property
    def dev_id(self):
        return None


# class OpGen(object):
#     def __init__(self, op, name):
#         self.op = op
#         self.name = name
#         self.cache = dict()

#     def __call__(self, *args, **kwargs):
#         if self.name not in self.cache:
#             # register operator
#             self.cache[self.name] = self.register()
#         try:
#             # forward and backward
#             kwargs.pop('__input_type__')
#             return self.cache[self.name](*args, **kwargs)
#         except KeyError:
#             # only forward
#             return self.cache[self.name]()(*args, **kwargs)

#     def register(self):
#         def forward(self, *args, **kwargs):
#             inputs, pars = get_in_data(op=self.op, *args, **kwargs)

#             self.in_data = inputs
#             dtype = self.in_data[0].dtype
#             self.req = ['write' for _ in self.in_data]
#             in_shape = get_in_shape(self.in_data)
#             out_shape = self.infer_shape(in_shape)[1]
#             self.out_data = [self.F.empty(s, dtype=dtype) for s in out_shape]
#             out = self._forward(*inputs)
#             if out is not None:
#                 if not isinstance(out, (list, tuple)):
#                     out = [out]
#                 for i, x in enumerate(out):
#                     self.assign(self.out_data[i], self.req[i], x)
#             if len(self.out_data) == 1:
#                 return self.out_data[0]
#             return self.out_data

#         def backward(self, out_grad=None, in_data=None, out_data=None, in_grad=None, req=None):

#             if in_data is not None:
#                 self.in_data = in_data
#             if out_data is not None:
#                 self.out_data = out_data

#             dtype = self.in_data[0].dtype

#             if in_grad is None:
#                 in_grad = [self.F.empty_like(d, dtype=dtype)
#                            for d in self.in_data]
#             else:
#                 if not isinstance(in_grad, (list, tuple)):
#                     in_grad = [in_grad]
#             self.in_grad = in_grad

#             if out_grad is None:
#                 out_grad = [self.F.ones_like(d, dtype=dtype)
#                             for d in self.out_data]
#             else:
#                 if not isinstance(out_grad, (list, tuple)):
#                     out_grad = [out_grad]
#             self.out_grad = out_grad

#             if req is None:
#                 self.req = ['write' for _ in self.in_data]
#             else:
#                 assert len(req) == len(self.in_data),\
#                     ValueError('len(req) should be %d' % len(self.in_data))
#                 self.req = req
#             out = self._backward(*out_grad)
#             if out is not None:
#                 if not isinstance(out, (list, tuple)):
#                     out = [out]
#                 num_inputs = len(get_varnames(self._forward))
#                 for i in range(num_inputs):
#                     self.assign(in_grad[i], self.req[i], out[i])
#             if len(in_grad) == 1:
#                 return in_grad[0]
#             return self.in_grad

#         np_op_dict = dict(
#             __call__=forward,
#             forward=forward,
#             backward=backward,
#             _forward=self.op.forward,
#             _backward=self.op.backward,
#             infer_shape=self.op.infer_shape,
#             assign=assign,
#             F=property(lambda self: np),
#             op=property(lambda dummy: self.op)
#         )
#         if hasattr(self.op, '__init__'):
#             np_op_dict['__init__'] = self.op.__init__
#         np_op_dict.update(INPUT_FUNCS)
#         np_op = type('_%s_NP_OP' % self.name,
#                      (self.op, object),
#                      np_op_dict)
#         return np_op


F = np



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
from antgo.framework.helper.cnn import ConvModule, DepthwiseSeparableConvModule
from antgo.framework.helper.base_module import BaseModule


class DarknetBottleneck(BaseModule):
    """The basic bottleneck block used in Darknet.

    Each ResBlock consists of two ConvModules and the input is added to the
    final output. Each ConvModule is composed of Conv, BN, and LeakyReLU.
    The first convLayer has filter size of 1x1 and the second one has the
    filter size of 3x3.

    Args:
        in_channels (int): The input channels of this Module.
        out_channels (int): The output channels of this Module.
        expansion (int): The kernel size of the convolution. Default: 0.5
        add_identity (bool): Whether to add identity to the out.
            Default: True
        use_depthwise (bool): Whether to use depthwise separable convolution.
            Default: False
        conv_cfg (dict): Config dict for convolution layer. Default: None,
            which means using conv2d.
        norm_cfg (dict): Config dict for normalization layer.
            Default: dict(type='BN').
        act_cfg (dict): Config dict for activation layer.
            Default: dict(type='Swish').
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 expansion=0.5,
                 add_identity=True,
                 use_depthwise=False,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN', momentum=0.03, eps=0.001),
                 act_cfg=dict(type='Swish'),
                 init_cfg=None):
        super().__init__(init_cfg)
        hidden_channels = int(out_channels * expansion)
        conv = DepthwiseSeparableConvModule if use_depthwise else ConvModule
        self.conv1 = ConvModule(
            in_channels,
            hidden_channels,
            1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)
        self.conv2 = conv(
            hidden_channels,
            out_channels,
            3,
            stride=1,
            padding=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)
        self.add_identity = \
            add_identity and in_channels == out_channels

    def forward(self, x):
        identity = x
        out = self.conv1(x)
        out = self.conv2(out)

        if self.add_identity:
            return out + identity
        else:
            return out


class CSPLayer(BaseModule):
    """Cross Stage Partial Layer.

    Args:
        in_channels (int): The input channels of the CSP layer.
        out_channels (int): The output channels of the CSP layer.
        expand_ratio (float): Ratio to adjust the number of channels of the
            hidden layer. Default: 0.5
        num_blocks (int): Number of blocks. Default: 1
        add_identity (bool): Whether to add identity in blocks.
            Default: True
        use_depthwise (bool): Whether to depthwise separable convolution in
            blocks. Default: False
        conv_cfg (dict, optional): Config dict for convolution layer.
            Default: None, which means using conv2d.
        norm_cfg (dict): Config dict for normalization layer.
            Default: dict(type='BN')
        act_cfg (dict): Config dict for activation layer.
            Default: dict(type='Swish')
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 expand_ratio=0.5,
                 num_blocks=1,
                 add_identity=True,
                 use_depthwise=False,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN', momentum=0.03, eps=0.001),
                 act_cfg=dict(type='Swish'),
                 init_cfg=None):
        super().__init__(init_cfg)
        mid_channels = int(out_channels * expand_ratio)
        self.main_conv = ConvModule(
            in_channels,
            mid_channels,
            1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)
        self.short_conv = ConvModule(
            in_channels,
            mid_channels,
            1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)
        self.final_conv = ConvModule(
            2 * mid_channels,
            out_channels,
            1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)

        self.blocks = nn.Sequential(*[
            DarknetBottleneck(
                mid_channels,
                mid_channels,
                1.0,
                add_identity,
                use_depthwise,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=act_cfg) for _ in range(num_blocks)
        ])

    def forward(self, x):
        x_short = self.short_conv(x)

        x_main = self.main_conv(x)
        x_main = self.blocks(x_main)

        x_final = torch.cat((x_main, x_short), dim=1)
        return self.final_conv(x_final)



from collections import defaultdict
import json
import os
import queue
from .build_hash import get_file_hash
from .build_path import get_virtual_dirname
from .build_dependant import get_dependant_graph

UPDATED_FLAG = '__updated_flag__'
HASH_FILENAME = 'antgo_op.hash.cache'
GRAPH_FILENAME = 'antgo_op.graph.cache'
HASH_FILE_BUFFER = defaultdict(dict)
GRAPH_FILE_BUFFER = defaultdict(dict)
FILE_CHANGED_STATE = dict()


def _read_hash_file(fname):
    global HASH_FILE_BUFFER
    '''read the file which stores the hash number of code files
       Json Format:
       dict(
        filename1=hashcode1,
        filename2=hashcode2,
       )
    '''
    data = HASH_FILE_BUFFER.get(fname, None)
    if data is not None:
        return data
    if os.path.exists(fname):
        with open(fname, 'r') as fin:
            data = json.load(fin)
    else:
        data = dict()
    HASH_FILE_BUFFER[fname] = data
    return data


def _write_hash_file(fname, data):
    with open(fname, 'w') as fout:
        json.dump(data, fout)


def _read_graph_file(fname):
    global GRAPH_FILE_BUFFER

    dirname = os.path.join('.temp', 'cache', os.path.basename(fname).split('.')[0])
    os.makedirs(dirname, exist_ok=True)
    fname = os.path.join(dirname, os.path.basename(fname))

    data = GRAPH_FILE_BUFFER.get(fname, None)
    if data is not None:
        return data
    if os.path.exists(fname):
        with open(fname, 'r') as fin:
            data = json.load(fin)
    else:
        data = dict()
    GRAPH_FILE_BUFFER[fname] = data
    return data


def _write_graph_file(fname, data):
    with open(fname, 'w') as fout:
        json.dump(data, fout)


def file_is_changed(fname):
    '''whether the file is changed'''
    global FILE_CHANGED_STATE
    changed = FILE_CHANGED_STATE.get(fname, None)
    if changed is not None:
        return changed
    if not os.path.exists(fname):
        return False
    fname = os.path.abspath(fname)
    new_hash = get_file_hash(fname)
    # dirname = get_virtual_dirname(os.path.dirname(fname))
    dirname = os.path.join('.temp', 'cache', os.path.basename(fname).split('.')[0])
    os.makedirs(dirname, exist_ok=True)
    hash_fname = os.path.join(dirname, HASH_FILENAME)

    def _get_file_old_hash(fname):
        if not os.path.exists(hash_fname):
            return None
        data = _read_hash_file(hash_fname)
        return data.get(fname, None)

    old_hash = _get_file_old_hash(fname)
    changed = new_hash != old_hash
    FILE_CHANGED_STATE[fname] = changed
    HASH_FILE_BUFFER[hash_fname][fname] = new_hash
    if changed:
        HASH_FILE_BUFFER[hash_fname][UPDATED_FLAG] = True
    return changed


def code_need_to_rebuild(source):
    '''whether the code file should be rebuilt'''
    source = os.path.abspath(source)
    dirname = get_virtual_dirname(os.path.dirname(source))
    graph_fname = os.path.join(dirname, GRAPH_FILENAME)
    graph_buf = _read_graph_file(graph_fname)
    graph = graph_buf.get(source, None)
    if graph is None or file_is_changed(source):
        graph = get_dependant_graph(source)
        graph_buf[UPDATED_FLAG] = True
        graph_buf[source] = graph
    # check graph
    change_state = dict()
    # find all changed file
    files = []
    for k, vs in graph.items():
        files.append(k)
        files.extend(vs)
    files = list(set(files))
    q = queue.Queue()
    for fname in files:
        changed = file_is_changed(fname)
        change_state[fname] = changed or None  # True or None
        q.put(fname)
    while not q.empty():
        fname = q.get()
        change_state[fname] = True
        for v in graph[fname]:
            if change_state[v] is None:
                q.put(v)
    for k in change_state.keys():
        if change_state[k] is None:
            change_state[k] = False
    changed = change_state[source]
    return changed


def save_hash_files():
    for fname in HASH_FILE_BUFFER.keys():
        try:
            v = HASH_FILE_BUFFER[fname]
            v.pop(UPDATED_FLAG)
            _write_hash_file(fname, v)
        except KeyError:
            continue


def save_graph_files():
    for fname in GRAPH_FILE_BUFFER.keys():
        try:
            v = GRAPH_FILE_BUFFER[fname]
            v.pop(UPDATED_FLAG)
            _write_graph_file(fname, v)
        except KeyError:
            continue


def save_latest_state():
    save_hash_files()
    save_graph_files()



# Copyright (c) OpenMMLab. All rights reserved.
import torch


def preprocess_panoptic_gt(gt_labels, gt_masks, gt_semantic_seg, num_things,
                           num_stuff):
    """Preprocess the ground truth for a image.

    Args:
        gt_labels (Tensor): Ground truth labels of each bbox,
            with shape (num_gts, ).
        gt_masks (BitmapMasks): Ground truth masks of each instances
            of a image, shape (num_gts, h, w).
        gt_semantic_seg (Tensor): Ground truth of semantic
            segmentation with the shape (1, h, w).
            [0, num_thing_class - 1] means things,
            [num_thing_class, num_class-1] means stuff,
            255 means VOID.
        target_shape (tuple[int]): Shape of output mask_preds.
            Resize the masks to shape of mask_preds.

    Returns:
        tuple: a tuple containing the following targets.

            - labels (Tensor): Ground truth class indices for a
                image, with shape (n, ), n is the sum of number
                of stuff type and number of instance in a image.
            - masks (Tensor): Ground truth mask for a image, with
                shape (n, h, w).
    """
    num_classes = num_things + num_stuff
    things_labels = gt_labels
    gt_semantic_seg = gt_semantic_seg.squeeze(0)

    things_masks = gt_masks.pad(gt_semantic_seg.shape[-2:], pad_val=0)\
        .to_tensor(dtype=torch.bool, device=gt_labels.device)

    semantic_labels = torch.unique(
        gt_semantic_seg,
        sorted=False,
        return_inverse=False,
        return_counts=False)
    stuff_masks_list = []
    stuff_labels_list = []
    for label in semantic_labels:
        if label < num_things or label >= num_classes:
            continue
        stuff_mask = gt_semantic_seg == label
        stuff_masks_list.append(stuff_mask)
        stuff_labels_list.append(label)

    if len(stuff_masks_list) > 0:
        stuff_masks = torch.stack(stuff_masks_list, dim=0)
        stuff_labels = torch.stack(stuff_labels_list, dim=0)
        labels = torch.cat([things_labels, stuff_labels], dim=0)
        masks = torch.cat([things_masks, stuff_masks], dim=0)
    else:
        labels = things_labels
        masks = things_masks

    masks = masks.long()
    return labels, masks



import inspect
import math
import random
import numbers
from numbers import Number
from typing import Sequence
import numpy as np
import torch
from ..builder import PIPELINES
import torchvision.transforms as transforms


def to_tensor(data):
    """Convert objects of various python types to :obj:`torch.Tensor`.

    Supported types are: :class:`numpy.ndarray`, :class:`torch.Tensor`,
    :class:`Sequence`, :class:`int` and :class:`float`.
    """
    if isinstance(data, torch.Tensor):
        return data
    elif isinstance(data, np.ndarray):
        return torch.from_numpy(data)
    elif isinstance(data, Sequence):
        return torch.tensor(data)
    elif isinstance(data, int):
        return torch.LongTensor([data])
    elif isinstance(data, float):
        return torch.FloatTensor([data])
    else:
        raise TypeError(
            f'Type {type(data)} cannot be converted to tensor.'
            'Supported types are: `numpy.ndarray`, `torch.Tensor`, '
            '`Sequence`, `int` and `float`')


@PIPELINES.register_module()
class IAnyToTensor(object):

    def __init__(self, keys=['image']):
        self.keys = keys

    def __call__(self, results):
        for key in self.keys:
            results[key] = to_tensor(results[key])
        return results

    def __repr__(self):
        return self.__class__.__name__ + f'(keys={self.keys})'


@PIPELINES.register_module()
class IImageToTensor(object):

    def __init__(self, keys=['image']):
        self.keys = keys

    def __call__(self, results):
        for key in self.keys:
            img = results[key]
            if len(img.shape) < 3:
                img = np.expand_dims(img, -1)
            results[key] = to_tensor(img.transpose(2, 0, 1))
        return results

    def __repr__(self):
        return self.__class__.__name__ + f'(keys={self.keys})'



from .accuracy import Accuracy, accuracy
from .ae_loss import AssociativeEmbeddingLoss
from .balanced_l1_loss import BalancedL1Loss, balanced_l1_loss
from .cross_entropy_loss import (CrossEntropyLoss, binary_cross_entropy,
                                 cross_entropy, mask_cross_entropy)
from .dice_loss import DiceLoss
from .focal_loss import FocalLoss, sigmoid_focal_loss
from .gaussian_focal_loss import GaussianFocalLoss
from .gfocal_loss import DistributionFocalLoss, QualityFocalLoss
from .ghm_loss import GHMC, GHMR
from .iou_loss import (BoundedIoULoss, CIoULoss, DIoULoss, GIoULoss, IoULoss,
                       bounded_iou_loss, iou_loss)
from .kd_loss import KnowledgeDistillationKLDivLoss
from .mse_loss import MSELoss, mse_loss
from .pisa_loss import carl_loss, isr_p
from .seesaw_loss import SeesawLoss
from .smooth_l1_loss import L1Loss, SmoothL1Loss, l1_loss, smooth_l1_loss
from .utils import reduce_loss, weight_reduce_loss, weighted_loss
from .varifocal_loss import VarifocalLoss

__all__ = [
    'accuracy', 'Accuracy', 'cross_entropy', 'binary_cross_entropy',
    'mask_cross_entropy', 'CrossEntropyLoss', 'sigmoid_focal_loss',
    'FocalLoss', 'smooth_l1_loss', 'SmoothL1Loss', 'balanced_l1_loss',
    'BalancedL1Loss', 'mse_loss', 'MSELoss', 'iou_loss', 'bounded_iou_loss',
    'IoULoss', 'BoundedIoULoss', 'GIoULoss', 'DIoULoss', 'CIoULoss', 'GHMC',
    'GHMR', 'reduce_loss', 'weight_reduce_loss', 'weighted_loss', 'L1Loss',
    'l1_loss', 'isr_p', 'carl_loss', 'AssociativeEmbeddingLoss',
    'GaussianFocalLoss', 'QualityFocalLoss', 'DistributionFocalLoss',
    'VarifocalLoss', 'KnowledgeDistillationKLDivLoss', 'SeesawLoss', 'DiceLoss'
]



from abc import ABCMeta, abstractmethod
from collections import OrderedDict
import numpy as np
from numpy.lib.function_base import insert
import torch
import torch.distributed as dist
from antgo.framework.helper.runner import BaseModule


class BaseDetector(BaseModule):
    """Base class for detectors."""

    def __init__(self, init_cfg=None):
        # super(BaseDetector, self).__init__(init_cfg)
        super().__init__()
        # self.fp16_enabled = False

    @property
    def with_neck(self):
        """bool: whether the detector has a neck"""
        return hasattr(self, 'neck') and self.neck is not None

    @property
    def with_bbox(self):
        """bool: whether the detector has a bbox head"""
        return ((hasattr(self, 'roi_head') and self.roi_head.with_bbox)
                or (hasattr(self, 'bbox_head') and self.bbox_head is not None))

    @property
    def with_mask(self):
        """bool: whether the detector has a mask head"""
        return ((hasattr(self, 'roi_head') and self.roi_head.with_mask)
                or (hasattr(self, 'mask_head') and self.mask_head is not None))

    @abstractmethod
    def extract_feat(self, imgs):
        """Extract features from images."""
        pass

    def extract_feats(self, imgs):
        """Extract features from multiple images.

        Args:
            imgs (list[torch.Tensor]): A list of images. The images are
                augmented from the same image but in different ways.

        Returns:
            list[torch.Tensor]: Features of different images
        """
        assert isinstance(imgs, list)
        return [self.extract_feat(img) for img in imgs]

    def forward_train(self, image, image_meta, **kwargs):
        """
        Args:
            image (list[Tensor]): List of tensors of shape (1, C, H, W).
                Typically these should be mean centered and std scaled.
            image_meta (list[dict]): List of image info dict where each dict
                has: 'img_shape', 'scale_factor', 'flip', and may also contain
                'filename', 'ori_shape', 'pad_shape', and 'img_norm_cfg'.
                For details on the values of these keys, see
                :class:`mmdet.datasets.pipelines.Collect`.
            kwargs (keyword arguments): Specific to concrete implementation.
        """
        pass

    @abstractmethod
    def simple_test(self, image, image_meta, **kwargs):
        pass

    def forward_test(self, image, image_meta, **kwargs):
        """
        Args:
            image (List[Tensor]): the outer list indicates test-time
                augmentations and inner Tensor should have a shape NxCxHxW,
                which contains all images in the batch.
            image_meta (List[List[dict]]): the outer list indicates test-time
                augs (multiscale, flip, etc.) and the inner list indicates
                images in a batch.
        """
        return self.simple_test(image, image_meta, **kwargs)

    def forward(self, image, image_meta, return_loss=True, **kwargs):
        """Calls either :func:`forward_train` or :func:`forward_test` depending
        on whether ``return_loss`` is ``True``.

        Note this setting will change the expected inputs. When
        ``return_loss=True``, img and img_meta are single-nested (i.e. Tensor
        and List[dict]), and when ``resturn_loss=False``, img and img_meta
        should be double nested (i.e.  List[Tensor], List[List[dict]]), with
        the outer list indicating test time augmentations.
        """
        if torch.onnx.is_in_onnx_export():
            assert len(image_meta) == 1
            return self.onnx_export(image[0], image_meta[0])

        if return_loss:
            return self.forward_train(image, image_meta, **kwargs)
        else:
            return self.forward_test(image, image_meta, **kwargs)

    def _parse_losses(self, losses):
        """Parse the raw outputs (losses) of the network.

        Args:
            losses (dict): Raw output of the network, which usually contain
                losses and other necessary information.

        Returns:
            tuple[Tensor, dict]: (loss, log_vars), loss is the loss tensor \
                which may be a weighted sum of all losses, log_vars contains \
                all the variables to be sent to the logger.
        """
        log_vars = OrderedDict()
        for loss_name, loss_value in losses.items():
            if isinstance(loss_value, torch.Tensor):
                log_vars[loss_name] = loss_value.mean()
            elif isinstance(loss_value, list):
                log_vars[loss_name] = sum(_loss.mean() for _loss in loss_value)
            else:
                raise TypeError(
                    f'{loss_name} is not a tensor or list of tensors')

        loss = sum(_value for _key, _value in log_vars.items()
                   if 'loss' in _key)

        # If the loss_vars has different length, GPUs will wait infinitely
        if dist.is_available() and dist.is_initialized():
            log_var_length = torch.tensor(len(log_vars), device=loss.device)
            dist.all_reduce(log_var_length)
            message = (f'rank {dist.get_rank()}' +
                       f' len(log_vars): {len(log_vars)}' + ' keys: ' +
                       ','.join(log_vars.keys()))
            assert log_var_length == len(log_vars) * dist.get_world_size(), \
                'loss log variables are different across GPUs!\n' + message

        log_vars['loss'] = loss
        for loss_name, loss_value in log_vars.items():
            # TODO, 仅用于统计作用，有时存在数值不稳定，不清楚为什么？
            # # reduce loss when distributed training
            # if dist.is_available() and dist.is_initialized():
            #     loss_value = loss_value.data.clone()
            #     dist.all_reduce(loss_value.div_(dist.get_world_size()))
            log_vars[loss_name] = loss_value.item()

        return loss, log_vars

    def train_step(self, data, optimizer, **kwargs):
        """The iteration step during training.

        This method defines an iteration step during training, except for the
        back propagation and optimizer updating, which are done in an optimizer
        hook. Note that in some complicated cases or models, the whole process
        including back propagation and optimizer updating is also defined in
        this method, such as GAN.

        Args:
            data (dict): The output of dataloader.
            optimizer (:obj:`torch.optim.Optimizer` | dict): The optimizer of
                runner is passed to ``train_step()``. This argument is unused
                and reserved.

        Returns:
            dict: It should contain at least 3 keys: ``loss``, ``log_vars``, \
                ``num_samples``.

                - ``loss`` is a tensor for back propagation, which can be a
                  weighted sum of multiple losses.
                - ``log_vars`` contains all the variables to be sent to the
                  logger.
                - ``num_samples`` indicates the batch size (when the model is
                  DDP, it means the batch size on each GPU), which is used for
                  averaging the logs.
        """
        losses =  None
        if type(data) == list or type(data) == tuple:
            losses = self(*data, **kwargs)
        else:
            losses = self(**data, **kwargs)
                
        loss, log_vars = self._parse_losses(losses)
        outputs = dict(
            loss=loss, log_vars=log_vars, num_samples=len(data['image_meta']))

        return outputs

    def val_step(self, data, optimizer=None, **kwargs):
        """The iteration step during validation.

        This method shares the same signature as :func:`train_step`, but used
        during val epochs. Note that the evaluation after training epochs is
        not implemented with this method, but an evaluation hook.
        """
        results = None
        if type(data) == list or type(data) == tuple:
            results = self(*data, **kwargs)
        else:
            results = self(**data, **kwargs)

        return results
        
    def onnx_export(self, image, image_meta):
        raise NotImplementedError(f'{self.__class__.__name__} does '
                                  f'not support ONNX EXPORT')



import os

highlight = [
    {'color': '#00BFFF','name': 'DeepSkyBlue'},
    {'color': '#00F5FF', 'name': 'Turquoise1'},
    {'color': '#7FFFD4', 'name': 'Aquamarine1'},
    {'color': '#54FF9F', 'name': 'SeaGreen1'},
    {'color': '#00FF00', 'name': 'Green1'},
    {'color': '#FFF68F', 'name': 'Khaki1'},
    {'color': '#FFFFE0', 'name': 'LightYellow1'},
    {'color': '#FFFF00', 'name': 'Yellow1'},
    {'color': '#FFD700', 'name': 'Gold1'},
    {'color': '#FFC1C1', 'name': 'RosyBrown1'},
    {'color': '#FF6A6A', 'name': 'IndianRed1'},
    {'color': '#FF8247', 'name': 'Sienna1'},
    {'color': '#FFD39B', 'name': 'Burlywood1'},
    {'color': '#FFA54F', 'name': 'Tan1'},
    {'color': '#FF3030', 'name': 'Firebrick1'},
    {'color': '#FF8C69', 'name': 'Salmon1'},
    {'color': '#FFA500', 'name': 'Orange1'},
    {'color': '#FF0000', 'name': 'Red1'},
    {'color': '#FF1493', 'name': 'DeepPink1'},
    {'color': '#FF00FF', 'name': 'Magenta1'},
    {'color': '#BF3EFF', 'name': 'DarkOrchid1'},
    {'color': '#9B30FF', 'name': 'Purple1'},
    {'color': '#00FF7F', 'name': 'SpringGreen'},
    {'color': '#FFFFE0', 'name': 'LightYellow'},
    {'color': '#FFDEAD', 'name': 'NavajoWhite1'},
    {'color': '#FFF0F5', 'name': 'LavenderBlush1'},
    {'color': '#836FFF', 'name': 'SlateBlue1'},
    {'color': '#63B8FF', 'name': 'SteelBlue1'},
    {'color': '#00BFFF', 'name': 'DeepSkyBlue1'},
    {'color': '#CAE1FF', 'name': 'LightSteelBlue1'},
    {'color': '#BFEFFF', 'name': 'LightBlue1'},
    {'color': '#E0FFFF', 'name': 'LightCyan1'}
]

dark=[
    {'color': '#0000FF','name': 'Blue'},
    {'color': '#00868B', 'name': 'Turquoise4'},
    {'color': '#458B74', 'name': 'Aquamarine4'},
    {'color': '#2E8B57', 'name': 'SeaGreen4'},
    {'color': '#008B00', 'name': 'Green4'},
    {'color': '#8B864E', 'name': 'Khaki4'},
    {'color': '#8B8B7A', 'name': 'LightYellow4'},
    {'color': '#8B8B00', 'name': 'Yellow4'},
    {'color': '#CDAD00', 'name': 'Gold3'},
    {'color': '#CD9B9B', 'name': 'RosyBrown3'},
    {'color': '#8B3A3A', 'name': 'IndianRed4'},
    {'color': '#8B4726', 'name': 'Sienna4'},
    {'color': '#CDAA7D', 'name': 'Burlywood3'},
    {'color': '#CD853F', 'name': 'Tan3'},
    {'color': '#8B1A1A', 'name': 'Firebrick4'},
    {'color': '#8B4C39', 'name': 'Salmon4'},
    {'color': '#8B5A00', 'name': 'Orange4'},
    {'color': '#8B0000', 'name': 'Red4'},
    {'color': '#8B0A50', 'name': 'DeepPink4'},
    {'color': '#8B008B', 'name': 'Magenta4'},
    {'color': '#68228B', 'name': 'DarkOrchid4'},
    {'color': '#551A8B', 'name': 'Purple4'},
    {'color': '#00FF00', 'name': 'Green'},
    {'color': '#FFFF00', 'name': 'Yellow'},
    {'color': '#CDB38B', 'name': 'NavajoWhite3'},
    {'color': '#CDC1C5', 'name': 'LavenderBlush3'},
    {'color': '#6959CD', 'name': 'SlateBlue3'},
    {'color': '#4F94CD', 'name': 'SteelBlue3'},
    {'color': '#009ACD', 'name': 'DeepSkyBlue3'},
    {'color': '#A2B5CD', 'name': 'LightSteelBlue3'},
    {'color': '#9AC0CD', 'name': 'LightBlue3'},
    {'color': '#B4CDCD', 'name': 'LightCyan3'}
]